#!/usr/bin/python
"""
    Parts of this source file are derived from code from Electrum
    (https://github.com/spesmilo/electrum), as of December 9, 2015.

    These parts are (c) 2015 by Thomas Voegtlin.  All changes are
    subject to the following copyright.
"""
"""
    Virtualchain
    ~~~~~
    copyright: (c) 2014-2015 by Halfmoon Labs, Inc.
    copyright: (c) 2016 by Blockstack.org

    This file is part of Virtualchain

    Virtualchain is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Virtualchain is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Virtualchain. If not, see <http://www.gnu.org/licenses/>.
"""


import socket
import os
import sys
import time
import logging

import protocoin
from protocoin.clients import *
from protocoin.serializers import *
from protocoin.fields import *
from protocoin.exceptions import *

from keys import version_byte as VERSION_BYTE

import bits

from ....lib import hashing

DEBUG = True

try:
    from ....lib.config import get_logger
except:
    def get_logger(name=None):
        """
        Get virtualchain's logger
        """

        level = logging.CRITICAL
        if DEBUG:
            logging.disable(logging.NOTSET)
            level = logging.DEBUG

        if name is None:
            name = "<unknown>"
            level = logging.CRITICAL

        log = logging.getLogger(name=name)
        log.setLevel( level )
        console = logging.StreamHandler()
        console.setLevel( level )
        log_format = ('[%(asctime)s] [%(levelname)s] [%(module)s:%(lineno)d] (' + str(os.getpid()) + ') %(message)s' if DEBUG else '%(message)s')
        formatter = logging.Formatter( log_format )
        console.setFormatter(formatter)
        log.propagate = False

        if len(log.handlers) > 0:
            for i in xrange(0, len(log.handlers)):
                log.handlers.pop(0)
        
        log.addHandler(console)
        return log


log = get_logger("virtualchain-bitcoin-spv")

BLOCK_HEADER_SIZE = 81

GENESIS_BLOCK_HASH = None
GENESIS_BLOCK_MERKLE_ROOT = None
USE_MAINNET = False
USE_TESTNET = False

GENESIS_BLOCK_HASH_MAINNET = "000000000019d6689c085ae165831e934ff763ae46a2a6c172b3f1b60a8ce26f"
GENESIS_BLOCK_MERKLE_ROOT_MAINNET = "4a5e1e4baab89f3a32518a88c31bc87f618f76673e2cc77ab2127b7afdeda33b"

GENESIS_BLOCK_HASH_TESTNET = "0f9188f13cb7b2c71f2a335e3a4fc328bf5beb436012afca590b1a11466e2206"
GENESIS_BLOCK_MERKLE_ROOT_TESTNET = "4a5e1e4baab89f3a32518a88c31bc87f618f76673e2cc77ab2127b7afdeda33b"

if VERSION_BYTE == 0:
    # mainnet
    log.debug("Using mainnet")
    USE_MAINNET = True
    GENESIS_BLOCK_HASH = GENESIS_BLOCK_HASH_MAINNET
    GENESIS_BLOCK_MERKLE_ROOT = GENESIS_BLOCK_MERKLE_ROOT_MAINNET

elif VERSION_BYTE == 111:
    # testnet
    log.debug("Using testnet/regtest")
    USE_TESTNET = True
    GENESIS_BLOCK_HASH = GENESIS_BLOCK_HASH_TESTNET
    GENESIS_BLOCK_MERKLE_ROOT = GENESIS_BLOCK_HASH_TESTNET
else:
    raise Exception("Unknown version byte %s" % VERSION_BYTE)

BLOCK_DIFFICULTY_CHUNK_SIZE = 2016
BLOCK_DIFFICULTY_INTERVAL = 14*24*60*60  # two weeks, in seconds


class BlockHash(SerializableMessage):
    """
    Block hash to request
    """
    def __init__(self):
        self.block_hash = None 

    def __repr__(self):
        return "<%s %s>" % (self.__class__.__name__, "%064x" % self.block_hash)


class GetHeaders(SerializableMessage):
    """
    getheaders message
    """
    command = "getheaders"

    def __init__(self):
        self.version = PROTOCOL_VERSION
        self.block_hashes = []
        self.hash_stop = 0

    def add_block_hash( self, block_hash ):
        """
        Append up to 2000 block hashes for which to get headers.
        """
        if len(self.block_hashes) > 2000:
            raise Exception("A getheaders request cannot have over 2000 block hashes")

        hash_num = int("0x" + block_hash, 16)
        
        bh = BlockHash()
        bh.block_hash = hash_num

        self.block_hashes.append( bh )
        self.hash_stop = hash_num

    def num_block_hashes( self ):
        """
        Get the number of block headers to request
        """
        return len(self.block_hashes)

    def __repr__(self):
        return "<%s block_hashes=[%s]>" % (self.__class__.__name__, ",".join([str(h) for h in self.block_hashes]))


class BlockHashSerializer( Serializer ):
    """
    Serialization class for a BlockHash
    """
    model_class = BlockHash
    block_hash = Hash()


class GetHeadersSerializer( Serializer ):
    """
    Serialization class for a GetHeaders
    """
    model_class = GetHeaders
    version = UInt32LEField()
    block_hashes = ListField(BlockHashSerializer)
    hash_stop = Hash()


# monkey-patch
protocoin.serializers.MESSAGE_MAPPING['getheaders'] = GetHeadersSerializer


class BlockHeaderClient( BitcoinBasicClient ):
    """
    Client to fetch and store block headers.
    """

    coin = None

    def __init__(self, socket, headers_path, first_block_hash, last_block_id ):

        if VERSION_BYTE == 0:
            self.coin = "bitcoin"
        else:
            self.coin = "bitcoin_testnet"

        super(BlockHeaderClient, self).__init__(socket)
        self.path = headers_path
        self.last_block_id = last_block_id
        self.finished = False
        self.verack = False
        self.first_block_hash = first_block_hash

    
    def loop_exit( self ):
        """
        Stop the loop
        """
        self.finished = True
        self.close_stream()


    def run( self ):
        """
        Interact with the blockchain peer,
        until we get a socket error or we
        exit the loop explicitly.
        Return True on success
        Raise on error
        """

        self.handshake()

        try:
            self.loop()
        except socket.error, se:
            if self.finished:
                return True
            else:
                raise


    def hash_to_string( self, hash_int ):
        return "%064x" % hash_int 


    def handle_headers( self, message_header, block_headers_message ):
        """
        Handle a 'headers' message.
        NOTE: we request headers in order, so we will expect to receive them in order here.
        Verify that we do so.
        """
        log.debug("handle headers (%s)" % len(block_headers_message.headers))

        block_headers = block_headers_message.headers

        serializer = BlockHeaderSerializer()

        # verify that the local header chain connects to this sequence
        current_height = SPVClient.height( self.path )
        if current_height is None:
            assert USE_TESTNET
            current_height = -1

        assert (current_height >= 0 and USE_MAINNET) or USE_TESTNET, "Invalid height %s" % current_height
    
        last_header = None

        if current_height >= 0:
            last_header = SPVClient.read_header( self.path, current_height )
            log.debug("Receive %s headers (%s to %s)" % (len(block_headers), current_height, current_height + len(block_headers)))

        else:
            # first testnet header
            log.debug("Receive %s testnet headers (%s to %s)" % (len(block_headers), current_height + 1, current_height + len(block_headers)))
            last_header = {
                "version": block_headers[0].version,
                "prev_block_hash": "%064x" % block_headers[0].prev_block,
                "merkle_root": "%064x" % block_headers[0].merkle_root,
                "timestamp": block_headers[0].timestamp,
                "bits": block_headers[0].bits,
                "nonce": block_headers[0].nonce,
                "hash": block_headers[0].calculate_hash()
            }

        if (USE_MAINNET or (USE_TESTNET and current_height >= 0)) and last_header['hash'] != self.hash_to_string(block_headers[0].prev_block):
            raise Exception("Received discontinuous block header at height %s: hash '%s' (expected '%s')" % \
                    (current_height,
                    self.hash_to_string(block_headers[0].prev_block),
                    last_header['hash'] ))

        header_start = 1
        if USE_TESTNET and current_height < 0:
            # save initial header
            header_start = 0

        # verify that this sequence of headers constitutes a hash chain 
        for i in xrange(header_start, len(block_headers)):
            prev_block_hash = self.hash_to_string(block_headers[i].prev_block)
            if i > 0 and prev_block_hash != block_headers[i-1].calculate_hash():
                raise Exception("Block '%s' is not continuous with block '%s'" % \
                        prev_block_hash,
                        block_headers[i-1].calculate_hash())

        if current_height < 0:
            # save the first header 
            if not os.path.exists(self.path):
                with open(self.path, "wb") as f:
                    block_header_serializer = BlockHeaderSerializer()
                    bin_data = block_header_serializer.serialize( block_headers[0] )
                    f.write( bin_data )

            # got all headers, including the first
            current_height = 0

        # insert into to local headers database
        next_block_id = current_height + 1
        for block_header in block_headers:
            with open(self.path, "r+") as f:

                # omit tx count 
                block_header.txns_count = 0
                bin_data = serializer.serialize( block_header )

                if len(bin_data) != BLOCK_HEADER_SIZE:
                    raise Exception("Block %s (%s) has %s-byte header" % (next_block_id, block_header.calculate_hash(), len(bin_data)))

                # NOTE: the fact that we use seek + write ensures that we can:
                # * restart synchronizing at any point
                # * allow multiple processes to work on the chain safely (even if they're duplicating effort)
                f.seek( BLOCK_HEADER_SIZE * next_block_id, os.SEEK_SET )
                f.write( bin_data )

                if SPVClient.height( self.path ) >= self.last_block_id - 1:
                    break

                next_block_id += 1
            
        current_block_id = SPVClient.height( self.path )
        if current_block_id >= self.last_block_id - 1:
            # got all headers
            self.loop_exit()
            return

        prev_block_header = SPVClient.read_header( self.path, current_block_id )
        prev_block_hash = prev_block_header['hash']
        self.send_getheaders( prev_block_hash )

    
    def send_getheaders( self, prev_block_hash ):
        """
        Request block headers from a particular block hash.
        Will receive up to 2000 blocks, starting with the block *after*
        the given block hash (prev_block_hash)
        """

        getheaders = GetHeaders()

        getheaders.add_block_hash( prev_block_hash )

        log.debug("send getheaders")
        self.send_message( getheaders )


    def handshake(self):
        """
        This method will implement the handshake of the
        Bitcoin protocol. It will send the Version message,
        and block until it receives a VerAck
        """
        log.debug("handshake (version %s)" % PROTOCOL_VERSION)
        version = Version()
        version.services = 0    # can't send blocks
        log.debug("send Version")
        self.send_message(version)


    def handle_version(self, message_header, message):
        """
        This method will handle the Version message and
        will send a VerAck message when it receives the
        Version message.

        :param message_header: The Version message header
        :param message: The Version message
        """
        log.debug("handle version")
        verack = VerAck()
        log.debug("send VerAck")
        self.send_message(verack)
        self.verack = True

        # begin!
        self.send_getheaders( self.first_block_hash )


    def handle_ping(self, message_header, message):
        """
        This method will handle the Ping message and then
        will answer every Ping message with a Pong message
        using the nonce received.

        :param message_header: The header of the Ping message
        :param message: The Ping message
        """
        log.debug("handle ping")
        pong = Pong()
        pong.nonce = message.nonce
        log.debug("send pong")
        self.send_message(pong)


class SPVClient(object):
    """
    Simplified Payment Verification client.
    Accesses locally-stored headers obtained by BlockHeaderClient
    to verify and synchronize them with the blockchain.
    """

    def __init__(self, path):
        SPVClient.init( path )


    @classmethod
    def init(cls, path):
        """
        Set up an SPV client.
        If the locally-stored headers do not exist, then 
        create a stub headers file with the genesis block information.
        """
        if not os.path.exists( path ):

            block_header_serializer = BlockHeaderSerializer()
            genesis_block_header = BlockHeader()

            if USE_MAINNET:
                # we know the mainnet block header
                # but we don't know the testnet/regtest block header
                genesis_block_header.version = 1
                genesis_block_header.prev_block = 0
                genesis_block_header.merkle_root = int(GENESIS_BLOCK_MERKLE_ROOT, 16 )
                genesis_block_header.timestamp = 1231006505
                genesis_block_header.bits = int( "1d00ffff", 16 )
                genesis_block_header.nonce = 2083236893
                genesis_block_header.txns_count = 0

                with open(path, "wb") as f:
                    bin_data = block_header_serializer.serialize( genesis_block_header )
                    f.write( bin_data )
            

    @classmethod
    def height(cls, path):
        """
        Get the locally-stored block height
        """
        if os.path.exists( path ):
            sb = os.stat( path )
            h = (sb.st_size / BLOCK_HEADER_SIZE) - 1
            return h
        else:
            return None


    @classmethod
    def read_header_at( cls, f ):
        """
        Given an open file-like object, read a block header
        from it and return it as a dict containing:
        * version (int)
        * prev_block_hash (hex str)
        * merkle_root (hex str)
        * timestamp (int)
        * bits (int)
        * nonce (ini)
        * hash (hex str)
        """
        header_parser = BlockHeaderSerializer()
        hdr = header_parser.deserialize( f )
        h = {}
        h['version'] = hdr.version
        h['prev_block_hash'] = "%064x" % hdr.prev_block
        h['merkle_root'] = "%064x" % hdr.merkle_root
        h['timestamp'] = hdr.timestamp
        h['bits'] = hdr.bits
        h['nonce'] = hdr.nonce
        h['hash'] = hdr.calculate_hash()
        return h


    @classmethod
    def load_header_chain( cls, chain_path ):
        """
        Load the header chain from disk.
        Each chain element will be a dictionary with:
        * 
        """

        header_parser = BlockHeaderSerializer()
        chain = []
        height = 0
        with open(chain_path, "rb") as f:

            h = SPVClient.read_header_at( f )
            h['block_height'] = height 

            height += 1
            chain.append(h)

        return chain


    @classmethod
    def read_header(cls, headers_path, block_height):
        """
        Get a block header at a particular height from disk.
        Return the header if found
        Return None if not.
        """
        if os.path.exists(headers_path):
    
            header_parser = BlockHeaderSerializer()
            sb = os.stat( headers_path )
            if sb.st_size < BLOCK_HEADER_SIZE * block_height:
                # beyond EOF 
                return None 

            with open( headers_path, "rb" ) as f:
                f.seek( block_height * BLOCK_HEADER_SIZE, os.SEEK_SET )
                hdr = SPVClient.read_header_at( f )

            return hdr
        else:
            return None


    @classmethod
    def get_target(cls, path, index, chain=None):
        """
        Calculate the target difficulty at a particular difficulty interval (index).
        Return (bits, target) on success
        """
        if chain is None:
            chain = []  # Do not use mutables as default values!

        max_target = 0x00000000FFFF0000000000000000000000000000000000000000000000000000
        if index == 0:
            return 0x1d00ffff, max_target

        first = SPVClient.read_header( path, (index-1)*BLOCK_DIFFICULTY_CHUNK_SIZE)
        last = SPVClient.read_header( path, index*BLOCK_DIFFICULTY_CHUNK_SIZE - 1)
        if last is None:
            for h in chain:
                if h.get('block_height') == index*BLOCK_DIFFICULTY_CHUNK_SIZE - 1:
                    last = h

        nActualTimespan = last.get('timestamp') - first.get('timestamp')
        nTargetTimespan = BLOCK_DIFFICULTY_INTERVAL
        nActualTimespan = max(nActualTimespan, nTargetTimespan/4)
        nActualTimespan = min(nActualTimespan, nTargetTimespan*4)

        bits = last.get('bits')
        # convert to bignum
        MM = 256*256*256
        a = bits%MM
        if a < 0x8000:
            a *= 256
        target = (a) * pow(2, 8 * (bits/MM - 3))

        # new target
        new_target = min( max_target, (target * nActualTimespan)/nTargetTimespan )

        # convert it to bits
        c = ("%064X"%new_target)[2:]
        i = 31
        while c[0:2]=="00":
            c = c[2:]
            i -= 1

        c = int('0x'+c[0:6],16)
        if c >= 0x800000:
            c /= 256
            i += 1

        new_bits = c + MM * i
        return new_bits, new_target

   
    @classmethod 
    def block_header_verify( cls, headers_path, block_id, block_hash, block_header ):
        """
        Given the block's numeric ID, its hash, and the bitcoind-returned block_data,
        use the SPV header chain to verify the block's integrity.

        block_header must be a dict with the following structure:
        * version: protocol version (int)
        * prevhash: previous block hash (hex str)
        * merkleroot: block Merkle root (hex str)
        * timestamp: UNIX time stamp (int)
        * bits: difficulty bits (hex str)
        * nonce: PoW nonce (int)
        * hash: block hash (hex str)
        (i.e. the format that the reference bitcoind returns via JSON RPC)

        Return True on success
        Return False on error
        """
        prev_header = cls.read_header( headers_path, block_id - 1 )
        prev_hash = prev_header['hash']
        return bits.block_header_verify( block_header, prev_hash, block_hash )


    @classmethod 
    def block_verify( cls, verified_block_header, block_txids ):
        """
        Given the block's verified header structure (see block_header_verify) and
        its list of transaction IDs (as hex strings), verify that the transaction IDs are legit.

        Return True on success
        Return False on error.
        """

        block_data = {
            'merkleroot': verified_block_header['merkleroot'],
            'tx': block_txids
        }

        return bits.block_verify( block_data )


    @classmethod 
    def tx_hash( cls, tx ):
        """
        Calculate the hash of a transction structure given by bitcoind
        """
        tx_hex = bits.btc_bitcoind_tx_serialize( tx )
        tx_hash = hashing.bin_double_sha256(tx_hex.decode('hex'))[::-1].encode('hex')
        return tx_hash


    @classmethod
    def tx_verify( cls, verified_block_txids, tx ):
        """
        Given the block's verified block txids, verify that a transaction is legit.
        @tx must be a dict with the following fields:
        * locktime: int
        * version: int
        * vin: list of dicts with:
           * vout: int,
           * hash: hex str
           * sequence: int (optional)
           * scriptSig: dict with:
              * hex: hex str
        * vout: list of dicts with:
           * value: float
           * scriptPubKey: dict with:
              * hex: hex str
        """
        tx_hash = cls.tx_hash( tx )
        return tx_hash in verified_block_txids


    @classmethod 
    def tx_index( cls, verified_block_txids, verified_tx ):
        """
        Given a block's verified block txids and a verified transaction, 
        find out where it is in the list of txids (i.e. what's its index)?
        """
        tx_hash = cls.tx_hash( verified_tx )
        return verified_block_txids.index( tx_hash )


    @classmethod 
    def block_header_index( cls, path, block_header ):
        """
        Given a block's serialized header, go and find out what its
        block ID is (if it is present at all).

        Return the >= 0 index on success
        Return -1 if not found.

        NOTE: this is slow
        """
        with open( path, "r" ) as f:
            chain_raw = f.read()

        for blk in xrange(0, len(chain_raw) / (BLOCK_HEADER_SIZE)):
            if chain_raw[blk * BLOCK_HEADER_SIZE : blk * BLOCK_HEADER_SIZE + BLOCK_HEADER_SIZE] == block_header:
                return blk

        return -1


    @classmethod
    def verify_header_chain(cls, path, chain=None):
        """
        Verify that a given chain of block headers
        has sufficient proof of work.
        """
        if chain is None:
            chain = SPVClient.load_header_chain( path )

        prev_header = chain[0]
        
        for i in xrange(1, len(chain)):
            header = chain[i]
            height = header.get('block_height')
            prev_hash = prev_header.get('hash')
            if prev_hash != header.get('prev_block_hash'):
                log.error("prev hash mismatch: %s vs %s" % (prev_hash, header.get('prev_block_hash')))
                return False

            bits, target = SPVClient.get_target( path, height/BLOCK_DIFFICULTY_CHUNK_SIZE, chain)
            if bits != header.get('bits'):
                log.error("bits mismatch: %s vs %s" % (bits, header.get('bits')))
                return False

            _hash = header.get('hash')
            if int('0x'+_hash, 16) > target:
                log.error("insufficient proof of work: %s vs target %s" % (int('0x'+_hash, 16), target))
                return False

            prev_header = header

        return True


    @classmethod
    def sync_header_chain(cls, path, bitcoind_server, last_block_id ):
        """
        Synchronize our local block headers up to the last block ID given.
        @last_block_id is *inclusive*
        @bitcoind_server is host:port or just host
        """
        current_block_id = SPVClient.height( path )
        if current_block_id is None:
            assert USE_TESTNET
            current_block_id = -1

        assert (current_block_id >= 0 and USE_MAINNET) or USE_TESTNET

        if current_block_id < last_block_id:
          
            if USE_MAINNET:
                log.debug("Synchronize %s to %s" % (current_block_id, last_block_id))
            else:
                log.debug("Synchronize testnet %s to %s" % (current_block_id + 1, last_block_id ))

            # need to sync
            if current_block_id >= 0:
                prev_block_header = SPVClient.read_header( path, current_block_id )
                prev_block_hash = prev_block_header['hash']

            else:
                # can only happen when in testnet
                prev_block_hash = GENESIS_BLOCK_HASH_TESTNET

            # connect 
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

            # timeout (10 min)
            sock.settimeout(600)

            bitcoind_port = 8333
            if ":" in bitcoind_server:
                p = bitcoind_server.split(":")
                bitcoind_server = p[0]
                bitcoind_port = int(p[1])

            log.debug("connect to %s:%s" % (bitcoind_server, bitcoind_port))
            sock.connect( (bitcoind_server, bitcoind_port) )

            client = BlockHeaderClient( sock, path, prev_block_hash, last_block_id )

            # get headers
            client.run()

            # verify headers
            if SPVClient.height(path) < last_block_id:
                raise Exception("Did not receive all headers up to %s (only got %s)" % (last_block_id, SPVClient.height(path)))

            # defensive: make sure it's *exactly* that many blocks 

            rc = SPVClient.verify_header_chain( path )
            if not rc:
                raise Exception("Failed to verify headers (stored in '%s')" % path)

        log.debug("synced headers from %s to %s in %s" % (current_block_id, last_block_id, path))
        return True


if __name__ == "__main__":
    # test synchonize headers 
    try:
        bitcoind_server = sys.argv[1]
        headers_path = sys.argv[2]
        height = int(sys.argv[3])
    except:
        print >> sys.stderr, "Usage: %s bitcoind_server headers_path blockchain_height" % sys.argv[0]
        sys.exit(0)

    log.setLevel(logging.DEBUG)
    SPVClient.init( headers_path )
    rc = SPVClient.sync_header_chain( headers_path, bitcoind_server, height )
    if rc:
        print "Headers are up to date with %s and seem to have sufficient proof-of-work" % height



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    Virtualchain
    ~~~~~
    copyright: (c) 2014-2015 by Halfmoon Labs, Inc.
    copyright: (c) 2016 by Blockstack.org
    
    This file is part of Virtualchain
    
    Virtualchain is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    
    Virtualchain is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with Virtualchain.  If not, see <http://www.gnu.org/licenses/>.
"""

OPCODE_VALUES = {
   "OP_0": 0, 
   "OP_PUSHDATA1": 76,  # 0x4c
   "OP_PUSHDATA2": 77,  # 0x4d
   "OP_PUSHDATA4": 78,  # 0x4e
   "OP_1NEGATE": 79, 
   "OP_RESERVED": 80, 
   "OP_1": 81, 
   "OP_2": 82, 
   "OP_3": 83, 
   "OP_4": 84, 
   "OP_5": 85, 
   "OP_6": 86, 
   "OP_7": 87, 
   "OP_8": 88, 
   "OP_9": 89, 
   "OP_10": 90, 
   "OP_11": 91, 
   "OP_12": 92, 
   "OP_13": 93, 
   "OP_14": 94, 
   "OP_15": 95, 
   "OP_16": 96, 
   "OP_NOP": 97, 
   "OP_VER": 98, 
   "OP_IF": 99, 
   "OP_NOTIF": 100, 
   "OP_VERIF": 101, 
   "OP_VERNOTIF": 102, 
   "OP_ELSE": 103, 
   "OP_ENDIF": 104, 
   "OP_VERIFY": 105,  # 0x69
   "OP_RETURN": 106,  # 0x6a
   "OP_TOALTSTACK": 107, 
   "OP_FROMALTSTACK": 108, 
   "OP_2DROP": 109, 
   "OP_2DUP": 110, 
   "OP_3DUP": 111, 
   "OP_2OVER": 112, 
   "OP_2ROT": 113, 
   "OP_2SWAP": 114, 
   "OP_IFDUP": 115, 
   "OP_DEPTH": 116, 
   "OP_DROP": 117, 
   "OP_DUP": 118, 
   "OP_NIP": 119, 
   "OP_OVER": 120, 
   "OP_PICK": 121, 
   "OP_ROLL": 122, 
   "OP_ROT": 123, 
   "OP_SWAP": 124, 
   "OP_TUCK": 125, 
   "OP_CAT": 126, 
   "OP_SUBSTR": 127, 
   "OP_LEFT": 128, 
   "OP_RIGHT": 129, 
   "OP_SIZE": 130, 
   "OP_INVERT": 131, 
   "OP_AND": 132, 
   "OP_OR": 133, 
   "OP_XOR": 134, 
   "OP_EQUAL": 135,  # 0x87
   "OP_EQUALVERIFY": 136,  # 0x88
   "OP_RESERVED1": 137, 
   "OP_RESERVED2": 138, 
   "OP_1ADD": 139, 
   "OP_1SUB": 140, 
   "OP_2MUL": 141, 
   "OP_2DIV": 142, 
   "OP_NEGATE": 143, 
   "OP_ABS": 144, 
   "OP_NOT": 145, 
   "OP_0NOTEQUAL": 146, 
   "OP_ADD": 147, 
   "OP_SUB": 148, 
   "OP_MUL": 149, 
   "OP_DIV": 150, 
   "OP_MOD": 151, 
   "OP_LSHIFT": 152, 
   "OP_RSHIFT": 153, 
   "OP_BOOLAND": 154, 
   "OP_BOOLOR": 155, 
   "OP_NUMEQUAL": 156, 
   "OP_NUMEQUALVERIFY": 157, 
   "OP_NUMNOTEQUAL": 158, 
   "OP_LESSTHAN": 159, 
   "OP_GREATERTHAN": 160, 
   "OP_LESSTHANOREQUAL": 161, 
   "OP_GREATERTHANOREQUAL": 162, 
   "OP_MIN": 163, 
   "OP_MAX": 164, 
   "OP_WITHIN": 165, 
   "OP_RIPEMD160": 166, 
   "OP_SHA1": 167, 
   "OP_SHA256": 168, 
   "OP_HASH160": 169,  # 0xa9
   "OP_HASH256": 170, 
   "OP_CODESEPARATOR": 171, 
   "OP_CHECKSIG": 172,  # 0xac
   "OP_CHECKSIGVERIFY": 173,  # 0xad
   "OP_CHECKMULTISIG": 174,  # 0xae
   "OP_CHECKMULTISIGVERIFY": 175,  # 0xaf
   "OP_NOP1": 176, 
   "OP_NOP2": 177, 
   "OP_NOP3": 178, 
   "OP_NOP4": 179, 
   "OP_NOP5": 180, 
   "OP_NOP6": 181, 
   "OP_NOP7": 182, 
   "OP_NOP8": 183, 
   "OP_NOP9": 184, 
   "OP_NOP10": 185, 
   "OP_PUBKEYHASH": 253,  # 0xfd
   "OP_PUBKEY": 254,  # 0xfe
   "OP_INVALIDOPCODE": 255
}

OPCODE_NAMES = dict([(v, k) for (k, v) in OPCODE_VALUES.items()])




#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    Virtualchain
    ~~~~~
    copyright: (c) 2014 by Halfmoon Labs, Inc.
    copyright: (c) 2015-2018 by Blockstack.org
    
    This file is part of Virtualchain
    
    Virtualchain is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    
    Virtualchain is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with Virtualchain.  If not, see <http://www.gnu.org/licenses/>.
"""

from bitcoin_blockchain.keys import btc_is_multisig, btc_is_multisig_address, \
        btc_is_multisig_script, btc_is_singlesig, btc_get_singlesig_privkey, \
        btc_is_singlesig_address, btc_get_privkey_address

def is_multisig(privkey_info, blockchain='bitcoin', **blockchain_opts):
    """
    Is the given private key bundle a multisig bundle?
    """
    if blockchain == 'bitcoin':
        return btc_is_multisig(privkey_info, **blockchain_opts)
    else:
        raise ValueError('Unknown blockchain "{}"'.format(blockchain))


def is_multisig_address(addr, blockchain='bitcoin', **blockchain_opts):
    """
    Is the given address a multisig address?
    """
    if blockchain == 'bitcoin':
        return btc_is_multisig_address(addr, **blockchain_opts)
    else:
        raise ValueError('Unknown blockchain "{}"'.format(blockchain))


def is_multisig_script(script, blockchain='bitcoin', **blockchain_opts):
    """
    Is the given script a multisig script?
    """
    if blockchain == 'bitcoin':
        return btc_is_multisig_script(script, **blockchain_opts)
    else:
        raise ValueError('Unknown blockchain "{}"'.format(blockchain))


def is_singlesig(privkey_info, blockchain='bitcoin', **blockchain_opts):
    """
    Is the given private key bundle a single-sig key bundle?
    """
    if blockchain == 'bitcoin':
        return btc_is_singlesig(privkey_info, **blockchain_opts)
    else:
        raise ValueError('Unknown blockchain "{}"'.format(blockchain))


def get_singlesig_privkey(privkey_info, blockchain='bitcoin', **blockchain_opts):
    """
    Given a private key bundle, get the (single) private key
    """
    if blockchain == 'bitcoin':
        return btc_get_singlesig_privkey(privkey_info, **blockchain_opts)
    else:
        raise ValueError('Unknown blockchain "{}"'.format(blockchain))


def is_singlesig_address(addr, blockchain='bitcoin', **blockchain_opts):
    """
    Is the given address a single-sig address?
    """
    if blockchain == 'bitcoin':
        return btc_is_singlesig_address(addr, **blockchain_opts)
    else:
        raise ValueError('Unknown blockchain "{}"'.format(blockchain))


def get_privkey_address(privkey_info, blockchain='bitcoin', **blockchain_opts):
    """
    Get the address from a private key bundle
    """
    if blockchain == 'bitcoin':
        return btc_get_privkey_address(privkey_info, **blockchain_opts)
    else:
        raise ValueError('Unknown blockchain "{}"'.format(blockchain))




#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    Virtualchain
    ~~~~~
    copyright: (c) 2014 by Halfmoon Labs, Inc.
    copyright: (c) 2015-2018 by Blockstack.org
    
    This file is part of Virtualchain
    
    Virtualchain is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    
    Virtualchain is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with Virtualchain.  If not, see <http://www.gnu.org/licenses/>.
"""
from bitcoin_blockchain import btc_make_payment_script, btc_make_data_script, btc_script_hex_to_address

def script_hex_to_address( script_hex, blockchain='bitcoin', **blockchain_opts):
    """
    High-level API call (meant to be blockchain agnostic)
    Examine a script (hex-encoded) and extract an address.
    Return the address on success
    Return None on error
    """
    if blockchain == 'bitcoin':
        return btc_script_hex_to_address(script_hex, **blockchain_opts)
    else:
        raise ValueError("Unknown blockchain '{}'".format(blockchain))


def make_payment_script(address, blockchain='bitcoin', **blockchain_opts):
    """
    High-level API call (meant to be blockchain agnostic)
    Make a pay-to-address script.
    """
    
    if blockchain == 'bitcoin':
        return btc_make_payment_script(address, **blockchain_opts)
    else:
        raise ValueError("Unknown blockchain '{}'".format(blockchain))


def make_data_script( data, blockchain='bitcoin', **blockchain_opts):
    """
    High-level API call (meant to be blockchain agnostic)
    Make a data-bearing transaction output.
    Data must be a hex string
    Returns a hex string.
    """
    if blockchain == 'bitcoin':
        return btc_make_data_script(data, **blockchain_opts)
    else:
        raise ValueError("Unknown blockchain '{}'".format(blockchain))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    Virtualchain
    ~~~~~
    copyright: (c) 2014-2015 by Halfmoon Labs, Inc.
    copyright: (c) 2016 by Blockstack.org
    
    This file is part of Virtualchain
    
    Virtualchain is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    
    Virtualchain is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with Virtualchain.  If not, see <http://www.gnu.org/licenses/>.
"""

import os
import re
import copy
import binascii
import simplejson
import keylib
from six import int2byte, b, integer_types
from decimal import *

from .keys import btc_script_hex_to_address, btc_is_singlesig, btc_is_multisig, \
        btc_script_deserialize, btc_script_serialize, btc_is_singlesig_segwit, btc_is_multisig_segwit, btc_get_singlesig_privkey, \
        btc_make_p2sh_p2wpkh_redeem_script, btc_make_p2sh_p2wsh_redeem_script

from .opcodes import *
from ....lib import encoding, ecdsalib, hashing, merkle
from ....lib.config import get_features

import traceback

from ....lib.config import get_logger
log = get_logger('virtualchain')

# signature modes
SIGHASH_ALL = 1
SIGHASH_NONE = 2
SIGHASH_SINGLE = 3
SIGHASH_ANYONECANPAY = 128

# bitcoin constants 
UINT_MAX = 4294967295


def read_as_int(ptr, buf, bytez):
    ptr[0] += bytez
    if ptr[0] > len(buf):
        raise ValueError("Invalid transaction: tried to parse {} bytes of {} offset in {}".format(bytez, ptr[0] - bytez, buf.encode('hex')))

    ret = encoding.decode( buf[ ptr[0]-bytez:ptr[0] ][::-1], 256 )
    return ret


def read_var_int(ptr, buf):
    ptr[0] += 1
    if ptr[0] > len(buf):
        raise ValueError("Invalid transaction: tried to parse {} byte of {} offset in {}".format(1, ptr[0] - 1, buf.encode('hex')))

    val = encoding.from_byte_to_int( buf[ ptr[0]-1 ] )
    if val < 253:
        return val

    ret = read_as_int(ptr, buf, pow(2, val - 252))
    return ret


def read_bytes(ptr, buf, bytez):
    ptr[0] += bytez
    if ptr[0] > len(buf):
        raise ValueError("Invalid transaction: tried to parse {} bytes of {} offset in {}".format(bytez, ptr[0] - bytez, buf.encode('hex')))

    ret = buf[ ptr[0]-bytez:ptr[0] ]
    return ret


def peek_bytes(ptr, buf, bytez):
    if ptr[0] + bytez > len(buf):
        raise ValueError("Invalid transaction: tried to parse {} bytes of {} offset in {}".format(bytez, ptr[0], buf.encode('hex')))

    ret = buf[ ptr[0]:ptr[0]+bytez ]
    return ret


def read_var_string(ptr, buf):
    size = read_var_int(ptr, buf)
    return read_bytes(ptr, buf, size)


def read_tx_body(ptr, tx):
    """
    Returns {'ins': [...], 'outs': [...]}
    """
    _obj = {"ins": [], "outs": [], 'locktime': None}

    # number of inputs
    ins = read_var_int(ptr, tx)

    # all inputs
    for i in range(ins):
        _obj["ins"].append({
            "outpoint": {
                "hash": read_bytes(ptr, tx, 32)[::-1],
                "index": read_as_int(ptr, tx, 4)
            },
            "script": read_var_string(ptr, tx),
            "sequence": read_as_int(ptr, tx, 4)
        })

    # number of outputs
    outs = read_var_int(ptr, tx)

    # all outputs
    for i in range(outs):
        _obj["outs"].append({
            "value": read_as_int(ptr, tx, 8),
            "script": read_var_string(ptr, tx)
        })

    return _obj


def read_tx_witnesses(ptr, tx, num_witnesses):
    """
    Returns an array of witness scripts.
    Each witness will be a bytestring (i.e. encoding the witness script)
    """
    witnesses = []
    for i in xrange(0, num_witnesses):

        witness_stack_len = read_var_int(ptr, tx)
        witness_stack = []

        for j in xrange(0, witness_stack_len):

            stack_item = read_var_string(ptr, tx)
            witness_stack.append(stack_item)
             
        witness_script = btc_witness_script_serialize(witness_stack).decode('hex')
        witnesses.append(witness_script)

    return witnesses


def make_var_string(string):
    """
    Make a var-string (a var-int with the length, concatenated with the data)
    Return the hex-encoded string
    """
    s = None
    if isinstance(string, str) and re.match('^[0-9a-fA-F]*$', string):
        # convert from hex to bin, safely
        s = binascii.unhexlify(string)
    else:
        s = string[:]

    buf = encoding.num_to_var_int(len(s)) + s
    return buf.encode('hex')


def _btc_witness_serialize_unit(unit):
    """
    Encode one item of a BTC witness script
    Return the encoded item (as a string)

    Returns a byte string with the encoded unit

    Based on code from pybitcointools (https://github.com/vbuterin/pybitcointools)
    by Vitalik Buterin
    """
    
    if isinstance(unit, int):
        # pass literal
        return encoding.from_int_to_byte(unit)

    elif unit is None:
        # None means OP_0
        return b'\x00'

    else:
        if len(unit) < 256:
            # length (1 byte) + payload
            return encoding.from_int_to_byte(len(unit)) + unit

        elif len(unit) < 65536:
            # length (2 bytes, big-endian) + payload
            return encoding.encode(len(unit), 256, 2)[::-1] + unit
        else:
            # length (4 bytes, big-endian) + payload
            return encoding.encode(len(unit), 256, 4)[::-1] + unit


def btc_witness_script_serialize(_stack):
    """
    Given a deserialized witness script stack (i.e. the input-specific witness, as an array of
    Nones, ints, and strings), turn it back into a hex-encoded script
    """
    stack = _stack
    if encoding.json_is_base(_stack, 16):
        # hex-to-bin all hex strings 
        stack = encoding.json_changebase(_stack, lambda x: binascii.unhexlify(x))

    return encoding.safe_hexlify(_btc_witness_serialize_unit(len(stack)) + ''.join(map(lambda stack_unit: _btc_witness_serialize_unit(stack_unit), stack)))


def btc_witness_script_deserialize(_script):
    """
    Given a hex-encoded serialized witness script, turn it into a witness stack
    (i.e. an array of Nones, ints, and strings)
    """

    script = None
    if isinstance(_script, str) and re.match('^[0-9a-fA-F]*$', _script):
        # convert from hex to bin, safely
        script = binascii.unhexlify(_script)
    else:
        script = _script[:]

    # pointer to byte offset in _script (as an array due to Python scoping rules)
    ptr = [0]

    witness_stack_len = read_var_int(ptr, script)
    witness_stack = []

    for _ in xrange(0, witness_stack_len):

        stack_item = read_var_string(ptr, script)
        witness_stack.append(stack_item)

    return witness_stack


def btc_tx_deserialize(_tx, **blockchain_opts):
    """
    Given a hex-encoded transaction, decode it into an object
    with the following structure:
    {
        ins: [
            {
                outpoint: { hash: ..., index: ... }, 
                script: ...,
                sequence: ...,
                witness_script: ...,        # not included if not segwit
            }, ...
        ]
        outs: [
            {
                value: ...,
                script: ...
            }, ...
        ],
        version: ..., 
        locktime: ...
    }
    
    Derived from pybitcointools (https://github.com/vbuterin/pybitcointools) written by Vitalik Buterin
    """

    tx = None
    if isinstance(_tx, str) and re.match('^[0-9a-fA-F]*$', _tx):
        # convert from hex to bin, safely
        tx = binascii.unhexlify(_tx)
    else:
        tx = _tx[:]

    # pointer to byte offset in _tx (as an array due to Python scoping rules)
    ptr = [0]
    
    # top-level tx
    obj = {"ins": [], "outs": [], 'version': None, 'locktime': None}

    # get version
    obj["version"] = read_as_int(ptr, tx, 4)

    # segwit? (bip143)
    # 5th byte will be 0 and 6th byte will be flags (nonzero) if so
    bip143 = peek_bytes(ptr, tx, 2)
    if ord(bip143[0]) == 0 and ord(bip143[1]) != 0:
        # segwit
        # consume marker
        read_bytes(ptr, tx, 2)

        # get the rest of the body
        body = read_tx_body(ptr, tx)
        obj['ins'] = body['ins']
        obj['outs'] = body['outs']
     
        # read witnesses for each input
        witness_scripts = read_tx_witnesses(ptr, tx, len(obj['ins']))
      
        if len(witness_scripts) != len(obj['ins']):
            raise ValueError('Invald number of witnesses in {}'.format(_tx))

        for i in xrange(0, len(witness_scripts)):
            obj['ins'][i]['witness_script'] = witness_scripts[i]
        
    else:
        # non-segwit
        body = read_tx_body(ptr, tx)
        obj['ins'] = body['ins']
        obj['outs'] = body['outs']

    # locktime
    obj["locktime"] = read_as_int(ptr, tx, 4)

    if not ptr[0] == len(tx):
        log.warning('Did not parse entire tx ({} bytes remaining)'.format(len(tx) - ptr[0]))

    # hexlify each byte field 
    obj = encoding.json_changebase(obj, lambda x: encoding.safe_hexlify(x))
    return obj


def btc_tx_serialize(_txobj):
    """
    Given a transaction dict returned by btc_tx_deserialize, convert it back into a
    hex-encoded byte string.

    Derived from code written by Vitalik Buterin in pybitcointools (https://github.com/vbuterin/pybitcointools)
    """
    
    # output buffer
    o = []
    txobj = None
    if encoding.json_is_base(_txobj, 16):
        # txobj is built from hex strings already.  deserialize them 
        txobj = encoding.json_changebase(_txobj, lambda x: binascii.unhexlify(x))
    else:
        txobj = copy.deepcopy(_txobj)

    # version
    o.append(encoding.encode(txobj["version"], 256, 4)[::-1])

    # do we have any witness scripts?
    have_witness = False
    for inp in txobj['ins']:
        if inp.has_key('witness_script') and len(inp['witness_script']) > 0:
            have_witness = True
            break

    if have_witness:
        # add segwit marker 
        o.append('\x00\x01')
    
    # number of inputs
    o.append(encoding.num_to_var_int(len(txobj["ins"])))

    # all inputs
    for inp in txobj["ins"]:
        # input tx hash
        o.append(inp["outpoint"]["hash"][::-1])

        # input tx outpoint
        o.append(encoding.encode(inp["outpoint"]["index"], 256, 4)[::-1])

        # input scriptsig
        script = inp.get('script')
        if not script:
            script = bytes()

        scriptsig = encoding.num_to_var_int(len(script)) + script
        o.append(scriptsig)

        # sequence
        o.append(encoding.encode(inp.get("sequence", UINT_MAX - 1), 256, 4)[::-1])

    # number of outputs
    o.append(encoding.num_to_var_int(len(txobj["outs"])))

    # all outputs
    for out in txobj["outs"]:
        # value
        o.append(encoding.encode(out["value"], 256, 8)[::-1])

        # scriptPubKey
        scriptpubkey = encoding.num_to_var_int(len(out['script'])) + out['script']
        o.append(scriptpubkey)

    # add witnesses 
    if have_witness:
        for inp in txobj['ins']:
            witness_script = inp.get('witness_script')
            if not witness_script:
                witness_script = '\x00'

            o.append(witness_script)

    # locktime
    o.append(encoding.encode(txobj["locktime"], 256, 4)[::-1])

    # full string
    ret = ''.join( encoding.json_changebase(o, lambda x: encoding.safe_hexlify(x)) )
    return ret


def btc_bitcoind_tx_is_coinbase( tx ):
    """
    Is a transaction a coinbase transaction?
    tx is a bitcoind-given transaction structure
    """
    for inp in tx['vin']:
        if 'coinbase' in inp.keys():
            return True 

    return False


def btc_bitcoind_tx_serialize( tx ):
     """
     Convert a *Bitcoind*-given transaction into its hex string.

     tx format is {'vin': [...], 'vout': [...], 'locktime': ..., 'version': ...},
     with the same formatting rules as getrawtransaction.
     
     (in particular, each value in vout is a Decimal, in BTC)
     """
     tx_ins = []
     tx_outs = []

     try:
         for inp in tx['vin']:
             next_inp = {
                "outpoint": {
                   "index": int(inp['vout']),
                   "hash": str(inp['txid'])
                }
             }
             if 'sequence' in inp:
                 next_inp['sequence'] = int(inp['sequence'])
             else:
                 next_inp['sequence'] = UINT_MAX

             if 'scriptSig' in inp:
                 next_inp['script'] = str(inp['scriptSig']['hex'])
             else:
                 next_inp['script'] = ""

             if 'txinwitness' in inp:
                 next_inp['witness_script'] = btc_witness_script_serialize(inp['txinwitness'])

             tx_ins.append(next_inp)

         for out in tx['vout']:
         
             assert out['value'] < 1000, "High transaction value\n%s" % simplejson.dumps(tx, indent=4, sort_keys=True)
             next_out = {
                'value': int(Decimal(out['value'] * 10**8)),
                'script': str(out['scriptPubKey']['hex'])
             }
             tx_outs.append(next_out)

         tx_fields = {
            "locktime": int(tx['locktime']),
            "version": int(tx['version']),
            "ins": tx_ins,
            "outs": tx_outs
         }

         tx_serialized = btc_tx_serialize( tx_fields )
         return str(tx_serialized)

     except KeyError, ke:
         if btc_bitcoind_tx_is_coinbase(tx) and 'hex' in tx.keys():
             tx_serialized = tx['hex']
             return str(tx_serialized)

         log.error("Key error in:\n%s" % simplejson.dumps(tx, indent=4, sort_keys=True))
         traceback.print_exc()
         raise ke


def btc_tx_is_segwit( tx_serialized ):
    """
    Is this serialized (hex-encoded) transaction a segwit transaction?
    """
    marker_offset = 4       # 5th byte is the marker byte
    flag_offset = 5         # 6th byte is the flag byte
    
    marker_byte_string = tx_serialized[2*marker_offset:2*(marker_offset+1)]
    flag_byte_string = tx_serialized[2*flag_offset:2*(flag_offset+1)]

    if marker_byte_string == '00' and flag_byte_string != '00':
        # segwit (per BIP144)
        return True
    else:
        return False


def btc_tx_witness_strip( tx_serialized ):
    """
    Strip the witness information from a serialized transaction
    """
    if not btc_tx_is_segwit(tx_serialized):
        # already strippped
        return tx_serialized
     
    tx = btc_tx_deserialize(tx_serialized)
    for inp in tx['ins']:
        del inp['witness_script']

    tx_stripped = btc_tx_serialize(tx)
    return tx_stripped


def btc_tx_get_hash( tx_serialized, hashcode=None ):
    """
    Make a transaction hash (txid) from a hex tx, optionally along with a sighash.
    This DOES NOT WORK for segwit transactions
    """
    if btc_tx_is_segwit(tx_serialized):
        raise ValueError('Segwit transaction: {}'.format(tx_serialized))

    tx_bin = binascii.unhexlify(tx_serialized)
    if hashcode:
        return binascii.hexlify( hashing.bin_double_sha256(tx_bin + encoding.encode(int(hashcode), 256, 4)[::-1]) )

    else:
        return binascii.hexlify( hashing.bin_double_sha256(tx_bin)[::-1] )


def btc_tx_script_to_asm( script_hex ):
    """
    Decode a script into assembler
    """
    if len(script_hex) == 0:
        return ""

    try:
        script_array = btc_script_deserialize(script_hex)
    except:
        log.error("Failed to convert '%s' to assembler" % script_hex)
        raise

    script_tokens = []
    for token in script_array:
        if token is None:
            token = 0

        token_name = None

        if type(token) in [int,long]:
            token_name = OPCODE_NAMES.get(token, None)
            if token_name is None:
                token_name = str(token)
        
        else:
            token_name = token

        script_tokens.append(token_name)

    return " ".join(script_tokens)


def btc_tx_output_has_data(output, **blockchain_opts):
    """
    Does this output have user data?
    @output must be an element from the 'outs' list in btc_tx_deserialize()
    """
    return btc_tx_output_script_has_data(output['script'], **blockchain_opts)


def btc_tx_output_script_has_data(output_script, **blockchain_opts):
    """
    Does a btc output script have data? i.e. is it an OP_RETURN?
    The script must be hex-encoded
    """
    if len(output_script) < 2:
        return False

    return int(output_script[0:2], 16) == OPCODE_VALUES['OP_RETURN']


def btc_tx_extend(partial_tx_hex, new_inputs, new_outputs, **blockchain_opts):
    """
    Given an unsigned serialized transaction, add more inputs and outputs to it.
    @new_inputs and @new_outputs will be virtualchain-formatted:
    * new_inputs[i] will have {'outpoint': {'index':..., 'hash':...}, 'script':..., 'witness_script': ...}
    * new_outputs[i] will have {'script':..., 'value':... (in fundamental units, e.g. satoshis!)}
    """

    # recover tx
    tx = btc_tx_deserialize(partial_tx_hex)
    tx_inputs, tx_outputs = tx['ins'], tx['outs']
    locktime, version = tx['locktime'], tx['version']

    tx_inputs += new_inputs
    tx_outputs += new_outputs

    new_tx = {
        'ins': tx_inputs,
        'outs': tx_outputs,
        'locktime': locktime,
        'version': version,
    }

    new_unsigned_tx = btc_tx_serialize(new_tx)
    return new_unsigned_tx


def btc_tx_der_encode_integer(r):
    """
    Return a DER-encoded integer

    Based on code from python-ecdsa (https://github.com/warner/python-ecdsa)
    by Brian Warner.  Subject to the MIT license.
    """
    # borrowed from python-ecdsa
    if r < 0:
        raise ValueError('cannot support negative numbers')

    h = ("%x" % r).encode()
    if len(h) % 2:
        h = b("0") + h
    s = binascii.unhexlify(h)
    num = s[0] if isinstance(s[0], integer_types) else ord(s[0])
    if num <= 0x7f:
        return b("\x02") + int2byte(len(s)) + s
    else:
        # DER integers are two's complement, so if the first byte is
        # 0x80-0xff then we need an extra 0x00 byte to prevent it from
        # looking negative.
        return b("\x02") + int2byte(len(s)+1) + b("\x00") + s


def btc_tx_der_encode_length(l):
    """
    Return a DER-encoded length field

    Based on code from python-ecdsa (https://github.com/warner/python-ecdsa)
    by Brian Warner.  Subject to the MIT license.
    """
    if l < 0:
        raise ValueError("length cannot be negative")

    if l < 0x80:
        return int2byte(l)
    s = ("%x" % l).encode()
    if len(s) % 2:
        s = b("0") + s
    s = binascii.unhexlify(s)
    llen = len(s)
    return int2byte(0x80 | llen) + s


def btc_tx_der_encode_sequence(*encoded_pieces):
    """
    Return a DER-encoded sequence

    Based on code from python-ecdsa (https://github.com/warner/python-ecdsa)
    by Brian Warner.  Subject to the MIT license.
    """
    # borrowed from python-ecdsa
    total_len = sum([len(p) for p in encoded_pieces])
    return b('\x30') + btc_tx_der_encode_length(total_len) + b('').join(encoded_pieces)


def btc_tx_der_encode_signature(r, s):
    """
    Return a DER-encoded signature as a 2-item sequence

    Based on code from python-ecdsa (https://github.com/warner/python-ecdsa)
    by Brian Warner.  Subject to the MIT license.
    """
    # borrowed from python-ecdsa
    return btc_tx_der_encode_sequence(btc_tx_der_encode_integer(r), btc_tx_der_encode_integer(s))


def btc_tx_sighash( tx, idx, script, hashcode=SIGHASH_ALL):
    """
    Calculate the sighash of a non-segwit transaction.

    If it's SIGHASH_NONE, then digest the inputs but no outputs 
    If it's SIGHASH_SINGLE, then digest all inputs and all outputs up to i (excluding values and scripts), and fully digest the ith input and output
    If it's (something) | SIGHASH_ANYONECANPAY, then only digest the ith input.
    
    Return the double-sha256 digest of the relevant fields.

    THIS DOES NOT WORK WITH SEGWIT OUTPUTS

    Adapted from https://github.com/vbuterin/pybitcointools, by Vitalik Buterin
    """

    txobj = btc_tx_deserialize(tx)

    idx = int(idx)
    hashcode = int(hashcode)
    newtx = copy.deepcopy(txobj)

    # remove all scriptsigs in all inputs, except for the ith input's scriptsig.
    # the other inputs will be 'partially signed', except for SIGHASH_ANYONECANPAY mode.
    for i in xrange(0, len(newtx['ins'])):
        newtx['ins'][i]["script"] = ''
        if i == idx:
            if newtx['ins'][i].has_key('witness_script') and newtx['ins'][i]['witness_script']:
                raise ValueError('this method does not handle segwit inputs')

        if newtx['ins'][i].has_key('witness_script'):
            del newtx['ins'][i]['witness_script']

    newtx["ins"][idx]["script"] = script

    if (hashcode & 0x1f) == SIGHASH_NONE:
        # don't care about the outputs with this signature
        newtx["outs"] = []
        for inp in newtx['ins']:
            inp['sequence'] = 0

    elif (hashcode & 0x1f) == SIGHASH_SINGLE:
        # only signing for this input.
        # all outputs after this input will not be signed.
        # all outputs before this input will be partially signed (but not their values or scripts)
        if len(newtx['ins']) > len(newtx['outs']):
            raise ValueError('invalid hash code: {} inputs but {} outputs'.format(len(newtx['ins']), len(newtx['outs'])))

        newtx["outs"] = newtx["outs"][:len(newtx["ins"])]
        for out in newtx["outs"][:len(newtx["ins"]) - 1]:
            out['value'] = 2**64 - 1
            out['script'] = ""

    elif (hashcode & SIGHASH_ANYONECANPAY) != 0:
        # only going to sign this specific input, and nothing else
        newtx["ins"] = [newtx["ins"][idx]]
   
    signing_tx = btc_tx_serialize(newtx)
    sighash = btc_tx_get_hash( signing_tx, hashcode )
    return sighash


def btc_tx_sighash_segwit(tx, i, prevout_amount, prevout_script, hashcode=SIGHASH_ALL):
    """
    Calculate the sighash for a segwit transaction, according to bip143
    """
    txobj = btc_tx_deserialize(tx)

    hash_prevouts = encoding.encode(0, 256, 32)
    hash_sequence = encoding.encode(0, 256, 32)
    hash_outputs = encoding.encode(0, 256, 32)

    if (hashcode & SIGHASH_ANYONECANPAY) == 0:
        prevouts = ''
        for inp in txobj['ins']:
            prevouts += hashing.reverse_hash(inp['outpoint']['hash'])
            prevouts += encoding.encode(inp['outpoint']['index'], 256, 4)[::-1].encode('hex')

        hash_prevouts = hashing.bin_double_sha256(prevouts.decode('hex'))

        # print 'prevouts: {}'.format(prevouts)

    if (hashcode & SIGHASH_ANYONECANPAY) == 0 and (hashcode & 0x1f) != SIGHASH_SINGLE and (hashcode & 0x1f) != SIGHASH_NONE:
        sequences = ''
        for inp in txobj['ins']:
            sequences += encoding.encode(inp['sequence'], 256, 4)[::-1].encode('hex')

        hash_sequence = hashing.bin_double_sha256(sequences.decode('hex'))

        # print 'sequences: {}'.format(sequences)

    if (hashcode & 0x1f) != SIGHASH_SINGLE and (hashcode & 0x1f) != SIGHASH_NONE:
        outputs = ''
        for out in txobj['outs']:
            outputs += encoding.encode(out['value'], 256, 8)[::-1].encode('hex')
            outputs += make_var_string(out['script'])

        hash_outputs = hashing.bin_double_sha256(outputs.decode('hex'))

        # print 'outputs: {}'.format(outputs)

    elif (hashcode & 0x1f) == SIGHASH_SINGLE and i < len(txobj['outs']):
        outputs = ''
        outputs += encoding.encode(txobj['outs'][i]['value'], 256, 8)[::-1].encode('hex')
        outputs += make_var_string(txobj['outs'][i]['script'])

        hash_outputs = hashing.bin_double_sha256(outputs.decode('hex'))

        # print 'outputs: {}'.format(outputs)

    # print 'hash_prevouts: {}'.format(hash_prevouts.encode('hex'))
    # print 'hash_sequence: {}'.format(hash_sequence.encode('hex'))
    # print 'hash_outputs: {}'.format(hash_outputs.encode('hex'))
    # print 'prevout_script: {}'.format(prevout_script)
    # print 'prevout_amount: {}'.format(prevout_amount)

    sighash_preimage = ''
    sighash_preimage += encoding.encode(txobj['version'], 256, 4)[::-1].encode('hex')
    sighash_preimage += hash_prevouts.encode('hex')
    sighash_preimage += hash_sequence.encode('hex')

    # this input's prevout, script, amount, and sequence
    sighash_preimage += hashing.reverse_hash(txobj['ins'][i]['outpoint']['hash'])
    sighash_preimage += encoding.encode(txobj['ins'][i]['outpoint']['index'], 256, 4)[::-1].encode('hex')
    sighash_preimage += make_var_string(prevout_script)
    sighash_preimage += encoding.encode(prevout_amount, 256, 8)[::-1].encode('hex')
    sighash_preimage += encoding.encode(txobj['ins'][i]['sequence'], 256, 4)[::-1].encode('hex')

    sighash_preimage += hash_outputs.encode('hex')
    sighash_preimage += encoding.encode(txobj['locktime'], 256, 4)[::-1].encode('hex')
    sighash_preimage += encoding.encode(hashcode, 256, 4)[::-1].encode('hex')

    sighash = hashing.bin_double_sha256(sighash_preimage.decode('hex')).encode('hex')

    # print 'sighash_preimage: {}'.format(sighash_preimage)
    # print 'sighash: {}'.format(sighash)

    return sighash
    

def btc_tx_make_input_signature(tx, idx, prevout_script, privkey_str, hashcode):
    """
    Sign a single input of a transaction, given the serialized tx,
    the input index, the output's scriptPubkey, and the hashcode.

    tx must be a hex-encoded string
    privkey_str must be a hex-encoded private key

    Return the hex signature.

    THIS DOES NOT WORK WITH SEGWIT TRANSACTIONS
    """
    if btc_tx_is_segwit(tx):
        raise ValueError('tried to use the standard sighash to sign a segwit transaction')

    pk = ecdsalib.ecdsa_private_key(str(privkey_str))
    priv = pk.to_hex()

    # get the parts of the tx we actually need to sign
    sighash = btc_tx_sighash(tx, idx, prevout_script, hashcode)
    # print 'non-segwit sighash: {}'.format(sighash)

    # sign using uncompressed private key
    pk_uncompressed_hex, pubk_uncompressed_hex = ecdsalib.get_uncompressed_private_and_public_keys(priv)
    sigb64 = ecdsalib.sign_digest( sighash, priv )

    # sanity check 
    # assert ecdsalib.verify_digest( txhash, pubk_uncompressed_hex, sigb64 )

    sig_r, sig_s = ecdsalib.decode_signature(sigb64)
    sig_bin = btc_tx_der_encode_signature(sig_r, sig_s)
    sig = sig_bin.encode('hex') + encoding.encode(hashcode, 16, 2)

    return sig


def btc_tx_make_input_signature_segwit(tx, idx, prevout_amount, prevout_script, privkey_str, hashcode):
    """
    Sign a single input of a transaction, given the serialized tx,
    the input index, the output's scriptPubkey, and the hashcode.

    tx must be a hex-encoded string
    privkey_str must be a hex-encoded private key

    Return the hex signature.
    """
    # always compressed
    if len(privkey_str) == 64:
        privkey_str += '01'

    pk = ecdsalib.ecdsa_private_key(str(privkey_str))
    pubk = pk.public_key()
    
    priv = pk.to_hex()

    # must always be compressed
    pub = keylib.key_formatting.compress(pubk.to_hex())
    sighash = btc_tx_sighash_segwit(tx, idx, prevout_amount, prevout_script, hashcode=hashcode)
    
    # sign using uncompressed private key
    # pk_uncompressed_hex, pubk_uncompressed_hex = ecdsalib.get_uncompressed_private_and_public_keys(priv)
    sigb64 = ecdsalib.sign_digest( sighash, priv )

    # sanity check 
    # assert ecdsalib.verify_digest( txhash, pubk_uncompressed_hex, sigb64 )

    sig_r, sig_s = ecdsalib.decode_signature(sigb64)
    sig_bin = btc_tx_der_encode_signature(sig_r, sig_s)
    sig = sig_bin.encode('hex') + encoding.encode(hashcode, 16, 2)

    # print 'segwit signature: {}'.format(sig)
    return sig


def btc_tx_sign_multisig(tx, idx, redeem_script, private_keys, hashcode=SIGHASH_ALL):
    """
    Sign a p2sh multisig input (not segwit!).

    @tx must be a hex-encoded tx

    Return the signed transaction
    """
    
    from .multisig import parse_multisig_redeemscript

    # sign in the right order.  map all possible public keys to their private key
    txobj = btc_tx_deserialize(str(tx))

    privs = {}
    for pk in private_keys:
        pubk = ecdsalib.ecdsa_private_key(pk).public_key().to_hex()

        compressed_pubkey = keylib.key_formatting.compress(pubk)
        uncompressed_pubkey = keylib.key_formatting.decompress(pubk)

        privs[compressed_pubkey] = pk
        privs[uncompressed_pubkey] = pk

    m, public_keys = parse_multisig_redeemscript(str(redeem_script))

    used_keys, sigs = [], []
    for public_key in public_keys:
        if public_key not in privs:
            continue

        if len(used_keys) == m:
            break

        if public_key in used_keys:
            raise ValueError('Tried to reuse key in redeem script: {}'.format(public_key))

        pk_str = privs[public_key]
        used_keys.append(public_key)

        sig = btc_tx_make_input_signature(tx, idx, redeem_script, pk_str, hashcode)
        sigs.append(sig)

    if len(used_keys) != m:
        raise ValueError('Missing private keys (used {}, required {})'.format(len(used_keys), m))

    txobj["ins"][idx]["script"] = btc_script_serialize([None] + sigs + [redeem_script])
    return btc_tx_serialize(txobj)


def btc_tx_sign_multisig_segwit(tx, idx, prevout_amount, witness_script, private_keys, hashcode=SIGHASH_ALL, hashcodes=None, native=False):
    """
    Sign a native p2wsh or p2sh-p2wsh multisig input.

    @tx must be a hex-encoded tx

    Return the signed transaction
    """

    from .multisig import parse_multisig_redeemscript

    if hashcodes is None:
        hashcodes = [hashcode] * len(private_keys)

    txobj = btc_tx_deserialize(str(tx))
    privs = {}
    for pk in private_keys:
        pubk = ecdsalib.ecdsa_private_key(pk).public_key().to_hex()

        compressed_pubkey = keylib.key_formatting.compress(pubk)
        privs[compressed_pubkey] = pk

    m, public_keys = parse_multisig_redeemscript(witness_script)

    used_keys, sigs = [], []
    for i, public_key in enumerate(public_keys):
        if public_key not in privs:
            continue

        if len(used_keys) == m:
            break

        if public_key in used_keys:
            raise ValueError('Tried to reuse key in witness script: {}'.format(public_key))

        pk_str = privs[public_key]
        used_keys.append(public_key)

        sig = btc_tx_make_input_signature_segwit(tx, idx, prevout_amount, witness_script, pk_str, hashcodes[i])
        sigs.append(sig)

        # print ''

    if len(used_keys) != m:
        raise ValueError('Missing private keys (used {}, required {})'.format(len(used_keys), m))
   
    if native: 
        # native p2wsh
        txobj['ins'][idx]['witness_script'] = btc_witness_script_serialize([None] + sigs + [witness_script]) 

        # print 'segwit multisig: native p2wsh: witness script {}'.format(txobj['ins'][idx]['witness_script'])

    else:
        # p2sh-p2wsh
        redeem_script = btc_make_p2sh_p2wsh_redeem_script(witness_script)
        txobj['ins'][idx]['script'] = redeem_script
        txobj['ins'][idx]['witness_script'] = btc_witness_script_serialize([None] + sigs + [witness_script]) 
        
        # print 'segwit multisig: p2sh p2wsh: witness script {}'.format(txobj['ins'][idx]['witness_script'])
        # print 'segwit multisig: p2sh p2wsh: redeem script {}'.format(txobj['ins'][idx]['script'])
    
    return btc_tx_serialize(txobj)


def btc_tx_sign(tx_hex, idx, prevout_script, prevout_amount, private_key_info, scriptsig_type, hashcode=SIGHASH_ALL, hashcodes=None, redeem_script=None, witness_script=None):
    """
    Insert a scriptsig for an input that will later be spent by a p2pkh, p2pk, or p2sh scriptPubkey.

    @private_key_info is either a single private key, or a dict with 'redeem_script' and 'private_keys' defined.

    @redeem_script, if given, must NOT start with the varint encoding its length.
    However, it must otherwise be a hex string

    Return the transaction with the @idx'th scriptSig filled in.
    """
    new_tx = None
    
    # print 'sign input {} as {}'.format(idx, scriptsig_type)

    if scriptsig_type in ['p2pkh', 'p2pk']:
        if not btc_is_singlesig(private_key_info):
            raise ValueError('Need only one private key for {}'.format(scriptsig_type))

        pk = ecdsalib.ecdsa_private_key(str(private_key_info))
        pubk = pk.public_key()
        pub = pubk.to_hex()

        sig = btc_tx_make_input_signature(tx_hex, idx, prevout_script, private_key_info, hashcode)

        # print 'non-segwit sig: {}'.format(sig)
        # print 'non-segwit pubk: {}'.format(pub)

        # NOTE: sig and pub need to be hex-encoded
        txobj = btc_tx_deserialize(str(tx_hex))

        if scriptsig_type == 'p2pkh':
            # scriptSig + scriptPubkey is <signature> <pubkey> OP_DUP OP_HASH160 <pubkeyhash> OP_EQUALVERIFY OP_CHECKSIG
            txobj['ins'][idx]['script'] = btc_script_serialize([sig, pub])

        else:
            # p2pk
            # scriptSig + scriptPubkey is <signature> <pubkey> OP_CHECKSIG
            txobj['ins'][idx]['script'] = btc_script_serialize([sig])

        new_tx = btc_tx_serialize(txobj)

    elif scriptsig_type == 'p2wpkh' or scriptsig_type == 'p2sh-p2wpkh':
        # must be a segwit singlesig bundle
        if not btc_is_singlesig_segwit(private_key_info):
            raise ValueError('Keys are not for p2wpkh or p2sh-p2wpkh')

        privkey_str = str(btc_get_singlesig_privkey(private_key_info))
        pk = ecdsalib.ecdsa_private_key(privkey_str)
        pubk = pk.public_key()
        pub = keylib.key_formatting.compress(pubk.to_hex())

        # special bip141 rule: this is always 0x1976a914{20-byte-pubkey-hash}88ac
        prevout_script_sighash = '76a914' + hashing.bin_hash160(pub.decode('hex')).encode('hex') + '88ac'

        sig = btc_tx_make_input_signature_segwit(tx_hex, idx, prevout_amount, prevout_script_sighash, privkey_str, hashcode)

        txobj = btc_tx_deserialize(str(tx_hex))
        txobj['ins'][idx]['witness_script'] = btc_witness_script_serialize([sig, pub])

        if scriptsig_type == 'p2wpkh':
            # native
            # NOTE: sig and pub need to be hex-encoded
            txobj['ins'][idx]['script'] = ''

            if redeem_script:
                # goes in scriptSig 
                txobj['ins'][idx]['script'] = btc_script_serialize([redeem_script])

        else:
            # p2sh-p2wpkh
            redeem_script = btc_make_p2sh_p2wpkh_redeem_script(pub)
            txobj['ins'][idx]['script'] = redeem_script

            # print 'scriptsig {} from {} is {}'.format(pub, privkey_str, txobj['ins'][idx]['script'])

        new_tx = btc_tx_serialize(txobj)

        # print 'scriptsig type: {}'.format(scriptsig_type)
        # print 'segwit scriptsig: {}'.format(txobj['ins'][idx]['script'])
        # print 'segwit witness script: {}'.format(txobj['ins'][idx]['witness_script'])

    elif scriptsig_type == 'p2wsh' or scriptsig_type == 'p2sh-p2wsh':
        # only support p2wsh for multisig purposes at this time
        if not btc_is_multisig_segwit(private_key_info):
            raise ValueError('p2wsh requires a multisig key bundle')

        native = None
        if scriptsig_type == 'p2wsh':
            native = True
        else:
            native = False

        new_tx = btc_tx_sign_multisig_segwit(tx_hex, idx, prevout_amount, private_key_info['redeem_script'], private_key_info['private_keys'], hashcode=hashcode, hashcodes=hashcodes, native=native)
       
        txobj = btc_tx_deserialize(new_tx)

        # print 'segwit scriptsig: {}'.format(txobj['ins'][idx]['script'])
        # print 'segwit witness script: {}'.format(txobj['ins'][idx]['witness_script'])

    elif scriptsig_type == 'p2sh':
    
        if not redeem_script:
            # p2sh multisig
            if not btc_is_multisig(private_key_info):
                raise ValueError('No redeem script given, and not a multisig key bundle')

            new_tx = btc_tx_sign_multisig(tx_hex, idx, private_key_info['redeem_script'], private_key_info['private_keys'], hashcode=hashcode)

        else:
            # NOTE: sig and pub need to be hex-encoded
            txobj = btc_tx_deserialize(str(tx_hex))

            # scriptSig + scriptPubkey is <redeem script> OP_HASH160 <script hash> OP_EQUAL
            txobj['ins'][idx]['script'] = btc_script_serialize([redeem_script]) 
            new_tx = btc_tx_serialize(txobj)

    else:
        raise ValueError("Unknown script type {}".format(scriptsig_type))

    return new_tx


def btc_script_classify(scriptpubkey, private_key_info=None):
    """
    Classify a scriptpubkey, optionally also using the private key info that will generate the corresponding scriptsig/witness
    Return None if not known (nonstandard)
    """
    if scriptpubkey.startswith("76a914") and scriptpubkey.endswith("88ac") and len(scriptpubkey) == 50:
        return 'p2pkh'

    elif scriptpubkey.startswith("a914") and scriptpubkey.endswith("87") and len(scriptpubkey) == 46:
        # maybe p2sh-p2wpkh or p2sh-p2wsh?
        if private_key_info:
            if btc_is_singlesig_segwit(private_key_info):
                return 'p2sh-p2wpkh'
            elif btc_is_multisig_segwit(private_key_info):
                return 'p2sh-p2wsh'

        return 'p2sh'

    elif scriptpubkey.startswith('00') and len(scriptpubkey) == 44:
        return 'p2wpkh'

    elif scriptpubkey.startswith('00') and len(scriptpubkey) == 66:
        return 'p2wsh'

    script_tokens = btc_script_deserialize(scriptpubkey)
    if len(script_tokens) == 0:
        return None

    if script_tokens[0] == OPCODE_VALUES['OP_RETURN']:
        return "nulldata"

    elif scriptpubkey.endswith("ae"):
        return "multisig"

    elif len(script_tokens) == 2 and script_tokens[-1] == OPCODE_VALUES["OP_CHECKSIG"]:
        return "p2pk"

    return None


def btc_privkey_scriptsig_classify(private_key_info):
    """
    What kind of scriptsig can this private key make?
    """
    if btc_is_singlesig(private_key_info):
        return 'p2pkh'

    if btc_is_multisig(private_key_info):
        return 'p2sh'

    if btc_is_singlesig_segwit(private_key_info):
        return 'p2sh-p2wpkh'

    if btc_is_multisig_segwit(private_key_info):
        return 'p2sh-p2wsh'

    return None


def btc_tx_sign_input(tx, idx, prevout_script, prevout_amount, private_key_info, hashcode=SIGHASH_ALL, hashcodes=None, segwit=None, scriptsig_type=None, redeem_script=None, witness_script=None, **blockchain_opts):
    """
    Sign a particular input in the given transaction.
    @private_key_info can either be a private key, or it can be a dict with 'redeem_script' and 'private_keys' defined

    Returns the tx with the signed input
    """
    if segwit is None:
        segwit = get_features('segwit')
    
    if scriptsig_type is None:
        scriptsig_type = btc_privkey_scriptsig_classify(private_key_info)
    
    if scriptsig_type in ['p2wpkh', 'p2wsh', 'p2sh-p2wpkh', 'p2sh-p2wsh'] and not segwit:
        raise ValueError("Segwit is not enabled, but {} is a segwit scriptsig type".format(prevout_script))

    return btc_tx_sign(tx, idx, prevout_script, prevout_amount, private_key_info, scriptsig_type, hashcode=hashcode, hashcodes=hashcodes, redeem_script=redeem_script, witness_script=witness_script)


def btc_tx_sign_all_unsigned_inputs(private_key_info, prev_outputs, unsigned_tx_hex, scriptsig_type=None, segwit=None, **blockchain_opts):
    """
    Sign all unsigned inputs with a given key.
    Use the given outputs to fund them.

    @private_key_info: either a hex private key, or a dict with 'private_keys' and 'redeem_script'
    defined as keys.
    @prev_outputs: a list of {'out_script': xxx, 'value': xxx} that are in 1-to-1 correspondence with the unsigned inputs in the tx ('value' is in satoshis)
    @unsigned_hex_tx: hex transaction with unsigned inputs

    Returns: signed hex transaction
    """
    if segwit is None:
        segwit = get_features('segwit')
    
    txobj = btc_tx_deserialize(unsigned_tx_hex)
    inputs = txobj['ins']
    
    if scriptsig_type is None:
        scriptsig_type = btc_privkey_scriptsig_classify(private_key_info)

    tx_hex = unsigned_tx_hex
    prevout_index = 0
    
    # import json
    # print ''
    # print 'transaction:\n{}'.format(json.dumps(btc_tx_deserialize(unsigned_tx_hex), indent=4, sort_keys=True))
    # print 'prevouts:\n{}'.format(json.dumps(prev_outputs, indent=4, sort_keys=True))
    # print ''

    for i, inp in enumerate(inputs):
        do_witness_script = segwit
        if inp.has_key('witness_script'):
            do_witness_script = True

        elif segwit:
            # all inputs must receive a witness script, even if it's empty 
            inp['witness_script'] = ''

        if (inp['script'] and len(inp['script']) > 0) or (inp.has_key('witness_script') and len(inp['witness_script']) > 0):
            continue

        if prevout_index >= len(prev_outputs):
            raise ValueError("Not enough prev_outputs ({} given, {} more prev-outputs needed)".format(len(prev_outputs), len(inputs) - prevout_index))

        # tx with index i signed with privkey
        tx_hex = btc_tx_sign_input(str(unsigned_tx_hex), i, prev_outputs[prevout_index]['out_script'], prev_outputs[prevout_index]['value'], private_key_info, segwit=do_witness_script, scriptsig_type=scriptsig_type)
        unsigned_tx_hex = tx_hex
        prevout_index += 1

    return tx_hex


def block_header_serialize( inp ):
    """
    Given block header information, serialize it and return the hex hash.

    inp has:
    * version (int)
    * prevhash (str)
    * merkle_root (str)
    * timestamp (int)
    * bits (int)
    * nonce (int)

    Based on code from pybitcointools (https://github.com/vbuterin/pybitcointools)
    by Vitalik Buterin
    """
    
    # concatenate to form header
    o = encoding.encode(inp['version'], 256, 4)[::-1] + \
        inp['prevhash'].decode('hex')[::-1] + \
        inp['merkle_root'].decode('hex')[::-1] + \
        encoding.encode(inp['timestamp'], 256, 4)[::-1] + \
        encoding.encode(inp['bits'], 256, 4)[::-1] + \
        encoding.encode(inp['nonce'], 256, 4)[::-1]

    # get (reversed) hash
    h = hashing.bin_sha256(hashing.bin_sha256(o))[::-1].encode('hex')
    assert h == inp['hash'], (hashing.bin_sha256(o).encode('hex'), inp['hash'])

    return o.encode('hex')


def block_header_to_hex( block_data, prev_hash ):
    """
    Calculate the hex form of a block's header, given its getblock information from bitcoind.
    """
    header_info = {
       "version": block_data['version'],
       "prevhash": prev_hash,
       "merkle_root": block_data['merkleroot'],
       "timestamp": block_data['time'],
       "bits": int(block_data['bits'], 16),
       "nonce": block_data['nonce'],
       "hash": block_data['hash']
    }

    return block_header_serialize(header_info)


def block_header_verify( block_data, prev_hash, block_hash ):
    """
    Verify whether or not bitcoind's block header matches the hash we expect.
    """
    serialized_header = block_header_to_hex( block_data, prev_hash )
    candidate_hash_bin_reversed = hashing.bin_double_sha256(binascii.unhexlify(serialized_header))
    candidate_hash = binascii.hexlify( candidate_hash_bin_reversed[::-1] )

    return block_hash == candidate_hash


def block_verify( block_data ):
    """
    Given block data (a dict with 'merkleroot' hex string and 'tx' list of hex strings--i.e.
    a block compatible with bitcoind's getblock JSON RPC method), verify that the
    transactions are consistent.

    Return True on success
    Return False if not.
    """
    
    # verify block data txs 
    m = merkle.MerkleTree( block_data['tx'] )
    root_hash = str(m.root())

    return root_hash == str(block_data['merkleroot'])


def btc_tx_output_parse_script( scriptpubkey ):
    """
    Given the hex representation of a script,
    turn it into a nice, easy-to-read dict.
    The dict will have:
    * asm: the disassembled script as a string
    * hex: the raw hex (given as an argument)
    * type: the type of script

    Optionally, it will have:
    * addresses: a list of addresses the script represents (if applicable)
    * reqSigs: the number of required signatures (if applicable)
    """
    
    script_type = None
    reqSigs = None
    addresses = []
   
    script_type = btc_script_classify(scriptpubkey)
    script_tokens = btc_script_deserialize(scriptpubkey)

    if script_type in ['p2pkh']:
        script_type = "pubkeyhash"
        reqSigs = 1
        addr = btc_script_hex_to_address(scriptpubkey)
        if not addr:
            raise ValueError("Failed to parse scriptpubkey address")

        addresses = [addr]

    elif script_type in ['p2sh', 'p2sh-p2wpkh', 'p2sh-p2wsh']:
        script_type = "scripthash"
        reqsigs = 1
        addr = btc_script_hex_to_address(scriptpubkey)
        if not addr:
            raise ValueError("Failed to parse scriptpubkey address")

        addresses = [addr]

    elif script_type == 'p2pk':
        script_type = "pubkey"
        reqSigs = 1

    elif script_type is None:
        script_type = "nonstandard"

    ret = {
        "asm": btc_tx_script_to_asm(scriptpubkey),
        "hex": scriptpubkey,
        "type": script_type
    }

    if addresses is not None:
        ret['addresses'] = addresses

    if reqSigs is not None:
        ret['reqSigs'] = reqSigs

    # print 'parse script {}: {}'.format(scriptpubkey, ret)

    return ret





#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
     Virtualchain
     ~~~~~
     copyright: (c) 2014-2015 by Halfmoon Labs, Inc.
     copyright: (c) 2016 by Blockstack.org

     This file is part of Virtualchain

     Virtualchain is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     Virtualchain is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     You should have received a copy of the GNU General Public License
     along with Virtualchain.  If not, see <http://www.gnu.org/licenses/>.
"""

from .hashing import bin_double_sha256, bin_to_hex_reversed, hex_to_bin_reversed

def calculate_merkle_pairs(bin_hashes, hash_function=bin_double_sha256):
    """
    Calculate the parents of a row of a merkle tree.
    Takes in a list of binary hashes, returns a binary hash.

    The returned parents list is such that parents[i] == hash(bin_hashes[2*i] + bin_hashes[2*i+1]).
    """
    hashes = list(bin_hashes)

    # if there are an odd number of hashes, double up the last one
    if len(hashes) % 2 == 1:
        hashes.append(hashes[-1])

    new_hashes = []
    for i in range(0, len(hashes), 2):
        new_hashes.append(hash_function(hashes[i] + hashes[i+1]))

    return new_hashes


def verify_merkle_path(merkle_root_hex, serialized_path, leaf_hash_hex, hash_function=bin_double_sha256):
    """
    Verify a merkle path.  The given path is the path from two leaf nodes to the root itself.

    merkle_root_hex is a little-endian, hex-encoded hash.
    serialized_path is the serialized merkle path
    path_hex is a list of little-endian, hex-encoded hashes.

    Return True if the path is consistent with the merkle root.
    Return False if not.
    """
    
    merkle_root = hex_to_bin_reversed(merkle_root_hex)
    leaf_hash = hex_to_bin_reversed(leaf_hash_hex)

    path = MerkleTree.path_deserialize(serialized_path)
    path = [{'order': p['order'], 'hash': hex_to_bin_reversed(p['hash'])} for p in path]

    if len(path) == 0:
        raise ValueError("Empty path")

    cur_hash = leaf_hash
    for i in range(0, len(path)):
        if path[i]['order'] == 'l':
            # left sibling
            cur_hash = hash_function(path[i]['hash'] + cur_hash)
        elif path[i]['order'] == 'r':
            # right sibling
            cur_hash = hash_function(cur_hash + path[i]['hash'])
        elif path[i]['order'] == 'm':
            # merkle root
            assert len(path) == 1
            return cur_hash == path[i]['hash']

    return cur_hash == merkle_root


class MerkleTree(object):
    def __init__(self, hex_hashes, hash_function=bin_double_sha256):
        """
        Make a merkle tree.
        * hashes is a list of hex-encoded hashes.
        * hash_function is a callable that takes a string as input and outputs a hash.

        The Merkle tree hashes will be converted to big-endian.
        """
        if len(hex_hashes) == 0:
            raise ValueError("At least one hash is required.")

        self.rows = []
        
        # convert to binary big-endian
        hashes = [hex_to_bin_reversed(h) for h in hex_hashes]
        
        # build the rows of the merkle tree
        self.rows.append(hashes)
        while len(hashes) > 1:
            hashes = calculate_merkle_pairs(hashes, hash_function=hash_function)
            self.rows.append(hashes)


    def get(self, row_index, column_index):
        """
        Get the hash at a given row and column index.
        Raise ValueError on out-of-bounds
        """
        if row_index + 1 > len(self.rows):
            raise ValueError("There aren't that many rows.")

        row = self.rows(row_index)
        if column_index + 1 > len(row):
            raise ValueError("There aren't that many items in that row.")

        return row[column_index]


    def root(self):
        """
        Returns the hex-encoded root (little-endian)
        """
        # return the merkle root
        bin_merkle_root = self.rows[-1][0]
        return bin_to_hex_reversed(bin_merkle_root)


    @classmethod
    def path_serialize(cls, path):
        """
        Given a list of [{'hash': ..., 'order': ...}], serialize it to a string.
        """
        # make it into a netstring
        path_parts = ['{}-{}'.format(p['order'], p['hash']) for p in path]
        path_ns_parts = ['{}:{},'.format(len(pp), pp) for pp in path_parts]
        path_str = ''.join(path_ns_parts)
        return '{}:{},'.format(len(path_str), path_str)
        

    @classmethod
    def path_deserialize(cls, serialized_path):
        """
        Given a netstring of path parts, go and parse it back into [{'hash': ..., 'order': ...}]
        """
        def _chomp_netstring_payload(s):
            try:
                ns_len_str, ns_body = s.split(':', 1)
                ns_len = int(ns_len_str)
                assert ns_body[ns_len] == ','
                ns_payload = ns_body[:ns_len]
                return ns_payload, ns_body[ns_len+1:]
            except:
                raise ValueError("Invalid netstring '{}'".format(s))

        path_str, extra = _chomp_netstring_payload(serialized_path)
        if len(extra) > 0:
            raise ValueError("Danlging data in '{}'".format(serialized_path))

        path = []
        while True:
            path_part, path_str = _chomp_netstring_payload(path_str)
            try:
                order, hash_hex = path_part.split('-', 1)
                assert order in ['l', 'r', 'm']
                path.append({'order': order, 'hash': hash_hex})
            except:
                raise ValueError("Invalid path entry {}".format(path_part))

            if len(path_str) == 0:
                break

        return path


    def path(self, leaf_hash_hex, serialize=True):
        """
        Get the path (as a list of hashes) from the leaf hash to the root.
        leaf_hash_hex is hex-encoded, little-endian.

        The returned path will be a list of {'hash': hex-encoded little-endian hash, 'order': 'l' or 'r'}

        Raise ValueError if leaf_hash is not present in the tree.
        """
        
        leaf_hash = hex_to_bin_reversed(leaf_hash_hex)

        # sanity check
        found = False
        ri = None       # index into self.rows where leaf_hash occurs.  Note that self.rows[0] is the bottom (leaves) of the Merkle tree.
        for ri, row in enumerate(self.rows):
            found = found or (leaf_hash in row)
            if found:
                break

        if not found:
            raise ValueError("Hash {} is not present in Merkle tree {}".format(leaf_hash, self.root()))

        path = []
        cur_hash = leaf_hash
        for i in range(ri, len(self.rows)-1):
            # find sibling
            sibling = {}
            leaf_index = self.rows[i].index(cur_hash)
            if leaf_index % 2 == 0:
                # append left sibling
                sibling_hash = None
                if leaf_index+1 >= len(self.rows[i]):
                    # double-up the last hash
                    assert leaf_index+1 == len(self.rows[i]), 'leaf_index = {}, i = {}, len(rows[i]) = {}, rows[0] = {}'.format(leaf_index, i, len(self.rows[i]), ','.join(r.encode('hex') for r in self.rows[0]))
                    sibling_hash = self.rows[i][-1]
                else:
                    sibling_hash = self.rows[i][leaf_index+1]

                sibling['hash'] = bin_to_hex_reversed(sibling_hash)
                sibling['order'] = 'r'
            else:
                # append right sibling
                sibling['hash'] = bin_to_hex_reversed(self.rows[i][leaf_index-1])
                sibling['order'] = 'l'

            path.append(sibling)

            # find parent
            cur_hash = self.rows[i+1][leaf_index/2]
        
        if len(path) == 0:
            # single-node merkle tree
            path = [{'hash': bin_to_hex_reversed(self.rows[0][0]), 'order': 'm'}]

        if not serialize:
            return path
        else:
            return self.path_serialize(path)


if __name__ == "__main__":
    print 'test merkle tree'

    import os

    fixtures = [
        [os.urandom(32) for _ in range(0,32)],
        [os.urandom(32) for _ in range(0,31)],
        [os.urandom(32) for _ in range(0,1)],
        ['abcde' for _ in range(0,32)],
        ['abcde' for _ in range(0,31)],
        ['abcde' for _ in range(0,1)],
        [h.decode('hex') for h in ['02125585a9b812347c21c9f1827463ccb93a5096fb1f846b83652353fbb53418','f5c366f5aa9f21b4823ac167e8a062805f54bf99cb56be3fd1bd500bd5a20609','7358bf608d6ed1010cc86c1be5df4f773e7a2d6f00f3f3284bd742a55ea4a382','952dfe16fb9de054e291ef5d6dbacf381a93dffc12fdea74305fe3723016bf4b','a6c38dd08339595abd473c2351018d6d9ec968a44a53c6f06666141ec52568ae','aa950063768b7de1dbb0ae3bc3d872b3b7c8ac63defa02c7fe5e7cae93d11d23','c8d64f8ac60cbae1dd97cf373a4114097586fee63a8fed2e46add59eba7c8072','6a49bc047d9330cadeee4558da2a4da76aff1419d33f4b3d2800a55f26977fe4','b1eb9cae861e31b7177ce87fc3afde1618f69349dc1bf3e11deb86ea8586cf67','d8335aca79d8a0d20edd653905cc5d7a4a3e1986269d56b5949d08baad0046df','b58bfc1ab8c3557fb5de0014c58469f9ff6b68f407983fbe0b5387bd6485fc86']],
    ]

    for i, data in enumerate(fixtures):
        print '\nfixture {} ({} entries)\n'.format(i, len(data))

        data_hashes = [bin_to_hex_reversed(bin_double_sha256(d)) for d in data]

        mt = MerkleTree(data_hashes)
        mps = []
        for j, dh in enumerate(data_hashes):
            mp = mt.path(dh, serialize=False)
            mp_str = mt.path_serialize(mp)
            recombined_mp = mt.path_deserialize(mp_str)

            assert mp == recombined_mp

            mps.append(mp_str)
            
            print 'dh[{}] = {}'.format(j, dh)
            print 'path from {} to {} is {}'.format(dh, mt.root(), mp)
            print 'path from {} to {} is {}'.format(dh, mt.root(), mp_str)

        for i, mp in enumerate(mps):
            assert verify_merkle_path(mt.root(), mp, data_hashes[i]), 'failed to verify {} to {}'.format(data_hashes[i], mt.root())

        for i, mp in enumerate(mps):
            # corrupt a hash
            mp_obj = mt.path_deserialize(mp)
            mp_obj[-1]['hash'] = mp_obj[-1]['hash'][::-1]
            mp = mt.path_serialize(mp_obj)

            assert not verify_merkle_path(mt.root(), mp, data_hashes[i]), 'failed to verify {} to {}'.format(data_hashes[i], mt.root())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    Virtualchain
    ~~~~~
    copyright: (c) 2014 by Halfmoon Labs, Inc.
    copyright: (c) 2015 by Blockstack.org
    
    This file is part of Virtualchain
    
    Virtualchain is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    
    Virtualchain is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with Virtualchain.  If not, see <http://www.gnu.org/licenses/>.
"""

__version__ = '0.18.0.1'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    Virtualchain
    ~~~~~
    copyright: (c) 2014-2015 by Halfmoon Labs, Inc.
    copyright: (c) 2016 by Blockstack.org
    
    This file is part of Virtualchain
    
    Virtualchain is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    
    Virtualchain is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with Virtualchain.  If not, see <http://www.gnu.org/licenses/>.
"""

import traceback
import sys

from .opcodes import *
from .keys import *
from .bits import *
from ....lib import hashing
from ....lib.config import get_logger

import os
import binascii

log = get_logger('virtualchain')

def make_multisig_script( pubs, m ):
    """
    Make a multisig scriptSig/witness script, as a hex string
    """
    return btc_script_serialize( [m] + pubs + [len(pubs)] + [OPCODE_VALUES['OP_CHECKMULTISIG']] )
   

def make_multisig_address(pubs, m):
    """
    Make a multisig (p2sh) address, given the list of public keys (as hex strings) and the number required for validation
    """
    return btc_make_p2sh_address(make_multisig_script(pubs, m))


def make_multisig_segwit_address(pubs, m):
    """
    make an address for p2sh-p2wsh multisig
    """
    script = make_multisig_script(pubs, m)
    return make_multisig_segwit_address_from_witness_script(script)


def make_multisig_segwit_address_from_witness_script(script):
    """
    multisig witness script (p2sh-p2wsh) to address
    """
    script_hash = hashing.bin_sha256(script.decode('hex')).encode('hex')
    scriptsig_script = '0020' + script_hash
    addr = btc_make_p2sh_address(scriptsig_script)
    return addr


def make_multisig_info( m, pks, compressed=None ):
    """
    Make a multisig address and redeem script.
    @m of the given @pks must sign.

    Return {'address': p2sh address, 'redeem_script': redeem script, 'private_keys': private keys, 'segwit': False}
    * privkeys will be hex-encoded
    * redeem_script will be hex-encoded
    """

    pubs = []
    privkeys = []
    for pk in pks:
        priv = None
        if compressed in [True, False]:
            priv = BitcoinPrivateKey(pk, compressed=compressed)
        else:
            priv = BitcoinPrivateKey(pk)

        priv_hex = priv.to_hex()
        pub_hex = priv.public_key().to_hex()

        privkeys.append(priv_hex)
        pubs.append(pub_hex)

    script = make_multisig_script(pubs, m)
    addr = btc_make_p2sh_address(script)

    return {
        'address': addr,
        'redeem_script': script,
        'private_keys': privkeys,
        'segwit': False,
    }


def make_multisig_segwit_info( m, pks ):
    """
    Make either a p2sh-p2wpkh or p2sh-p2wsh
    redeem script and p2sh address.

    Return {'address': p2sh address, 'redeem_script': **the witness script**, 'private_keys': privkeys, 'segwit': True}
    * privkeys and redeem_script will be hex-encoded
    """
    pubs = []
    privkeys = []
    for pk in pks:
        priv = BitcoinPrivateKey(pk, compressed=True)
        priv_hex = priv.to_hex()
        pub_hex = priv.public_key().to_hex()

        privkeys.append(priv_hex)
        pubs.append(keylib.key_formatting.compress(pub_hex))

    script = None

    if len(pubs) == 1:
        if m != 1:
            raise ValueError("invalid m: len(pubkeys) == 1")

        # 1 pubkey means p2wpkh
        key_hash = hashing.bin_hash160(pubs[0].decode('hex')).encode('hex')
        script = '0014' + key_hash
        addr = btc_make_p2sh_address(script)

    else:
        # 2+ pubkeys means p2wsh 
        script = make_multisig_script(pubs, m)
        addr = make_multisig_segwit_address_from_witness_script(script)

    return {
        'address': addr,
        'redeem_script': script,
        'private_keys': privkeys,
        'segwit': True,
        'm': m
    }


def make_segwit_info(privkey):
    """
    Make a p2sh-p2wpkh wallet
    Return {'address': p2sh address, 'redeem_script': **the witness script**, 'private_keys': privkeys, 'segwit': True}
    """
    return make_multisig_segwit_info(1, [privkey])


def make_multisig_wallet( m, n ):
    """
    Create a bundle of information
    that can be used to generate an 
    m-of-n multisig scriptsig.
    """

    if m <= 1 and n <= 1:
        raise ValueError("Invalid multisig parameters")

    pks = []
    for i in xrange(0, n):
        pk = BitcoinPrivateKey(compressed=True).to_wif()
        pks.append(pk)

    return make_multisig_info( m, pks )


def make_segwit_info(privkey=None):
    """
    Create a bundle of information
    that can be used to generate
    a p2sh-p2wpkh transaction
    """

    if privkey is None:
        privkey = BitcoinPrivateKey(compressed=True).to_wif()

    return make_multisig_segwit_info(1, [privkey])


def make_multisig_segwit_wallet( m, n ):
    """
    Create a bundle of information
    that can be used to generate an
    m-of-n multisig witness script.
    """
    pks = []
    for i in xrange(0, n):
        pk = BitcoinPrivateKey(compressed=True).to_wif()
        pks.append(pk)

    return make_multisig_segwit_info(m, pks)


def parse_multisig_redeemscript( redeem_script_hex ):
    """
    Given a redeem script (as hex), extract multisig information.
    Return m, list of public keys on success
    Return (None, None)
    """
    script_parts = []
    redeem_script_hex = str(redeem_script_hex)

    try:
        script_parts = btc_script_deserialize(redeem_script_hex)
    except:
        if os.environ.get("BLOCKSTACK_TEST") == "1":
            traceback.print_exc()

        log.error("Invalid redeem script %s" % redeem_script_hex)
        return None, None

    try:
        assert len(script_parts) > 2
        assert script_parts[-1] == OPCODE_VALUES['OP_CHECKMULTISIG']
        script_parts.pop(-1)

        # get n
        n = script_parts.pop(-1)
        pubkeys = []

        # get m
        m = script_parts.pop(0)

        for i in xrange(0, n):
            pubk = script_parts.pop(0)
            
            # must be a public key
            BitcoinPublicKey(pubk)
            pubkeys.append(pubk)

        assert len(script_parts) == 0, "script_parts = %s" % script_parts
        return (m, pubkeys)
    except Exception, e:
        if os.environ.get("BLOCKSTACK_TEST") == "1":
            traceback.print_exc()

        log.error("Invalid redeem script %s (parses to %s)" % (redeem_script_hex, script_parts))
        return (None, None)


def parse_multisig_scriptsig( scriptsig_hex ):
    """
    Given a scriptsig (as hex), extract the signatures.
    Return list of signatures on success
    Return None on error
    """
    try:
        script_parts = btc_script_deserialize(scriptsig_hex)
    except:
        if os.environ.get("BLOCKSTACK_TEST") == "1":
            traceback.print_exc()

        return None

    # sanity check 
    return script_parts





# required callbacks
from .reference import get_virtual_chain_name, get_virtual_chain_version, get_first_block_id, get_db_state, db_parse, db_check, db_commit, db_save, db_serialize

# optional
try:
   from .reference import get_op_processing_order
except:
   def get_op_processing_order():
      return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    Virtualchain
    ~~~~~
    copyright: (c) 2014-2015 by Halfmoon Labs, Inc.
    copyright: (c) 2016 by Blockstack.org
    
    This file is part of Virtualchain
    
    Virtualchain is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    
    Virtualchain is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with Virtualchain.  If not, see <http://www.gnu.org/licenses/>.
"""

import keylib
import os
import binascii
import jsonschema
import re
from opcodes import *
from bech32 import *
from jsonschema import ValidationError

from ....lib import hashing, encoding, ecdsalib
from ....lib.config import get_features

MAX_DATA_LEN = 80       # 80 bytes per data output

OP_BASE58CHECK_PATTERN = r'^([123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz]+)$'
OP_ADDRESS_PATTERN = OP_BASE58CHECK_PATTERN
OP_PRIVKEY_PATTERN = OP_BASE58CHECK_PATTERN
OP_HEX_PATTERN = r'^([0-9a-fA-F]+)$'
OP_SCRIPT_PATTERN = OP_HEX_PATTERN

PRIVKEY_SINGLESIG_SCHEMA_WIF = {
    'type': 'string',
    'pattern': OP_PRIVKEY_PATTERN
}

PRIVKEY_SINGLESIG_SCHEMA_HEX = {
    'type': 'string',
    'pattern': OP_HEX_PATTERN
}

PRIVKEY_SINGLESIG_SCHEMA = {
    'anyOf': [
        PRIVKEY_SINGLESIG_SCHEMA_WIF,
        PRIVKEY_SINGLESIG_SCHEMA_HEX
    ],
}

PRIVKEY_MULTISIG_SCHEMA = {
    'type': 'object',
    'properties': {
        'address': {
            'type': 'string',
            'pattern': OP_ADDRESS_PATTERN,
        },
        'redeem_script': {
            'type': 'string',
            'pattern': OP_SCRIPT_PATTERN,
        },
        'private_keys': {
            'type': 'array',
            'items': PRIVKEY_SINGLESIG_SCHEMA
        },
        'segwit': {
            'type': 'boolean'
        },
    },
    'required': [
        'address',
        'redeem_script',
        'private_keys'
    ],
}


# depending on whether or not we're talking to 
# -testnet/-regtest or mainnet, determine which private
# and public key classes to use.

if os.environ.get("BLOCKSTACK_TESTNET", None) == "1" or os.environ.get("BLOCKSTACK_TESTNET3", None) == "1":

    version_byte = 111
    multisig_version_byte = 196

    # using testnet keys
    class TestnetPublicKey(keylib.ECPublicKey):
        _version_byte = 111

    class TestnetPrivateKey(keylib.ECPrivateKey):
        _pubkeyhash_version_byte = 111

    BitcoinPrivateKey = TestnetPrivateKey
    BitcoinPublicKey = TestnetPublicKey

else:

    version_byte = 0
    multisig_version_byte = 5

    # using mainnet keys
    BitcoinPrivateKey = keylib.ECPrivateKey
    BitcoinPublicKey = keylib.ECPublicKey


def bin_hash160_to_address(bin_hash160, version_byte=version_byte):
    # b58 addresses only!
    return keylib.b58check.b58check_encode(bin_hash160, version_byte=version_byte)


def hex_hash160_to_address(hash160, version_byte=version_byte):
    # b58 addresses only!
    return bin_hash160_to_address(binascii.unhexlify(hash160), version_byte=version_byte)


def address_to_bin_hash160(address):
    # b58 addresses only!
    return keylib.b58check.b58check_decode(address)


def address_to_hex_hash160(address):
    # b58 addresses only!
    return binascii.hexlify(address_to_bin_hash160(address))


def btc_script_to_hex(script):
    """ Parse the string representation of a script and return the hex version.
        Example: "OP_DUP OP_HASH160 c629...a6db OP_EQUALVERIFY OP_CHECKSIG"
    """

    hex_script = ''
    parts = script.split(' ')
    for part in parts:
        if part[0:3] == 'OP_':
            value = OPCODE_VALUES.get(part)
            if not value:
                raise ValueError("Unrecognized opcode {}".format(part))

            hex_script += "%0.2x" % value

        elif isinstance(part, (int)):
            hex_script += '%0.2x' % part

        elif hashing.is_hex(part):
            hex_script += '%0.2x' % hashing.count_bytes(part) + part

        else:
            raise Exception('Invalid script - only opcodes and hex characters allowed.')

    return hex_script


def btc_script_deserialize(script):
    """
    Given a script (hex or bin), decode it into its list of opcodes and data.
    Return a list of strings and ints.

    Based on code in pybitcointools (https://github.com/vbuterin/pybitcointools)
    by Vitalik Buterin
    """

    if isinstance(script, str) and re.match('^[0-9a-fA-F]*$', script):
       script = binascii.unhexlify(script)

    # output buffer
    out = []
    pos = 0

    while pos < len(script):
        # next script op...
        code = encoding.from_byte_to_int(script[pos])

        if code == 0:
            # empty (OP_0)
            out.append(None)
            pos += 1

        elif code <= 75:
            # numeric constant 
            out.append(script[pos+1:pos+1+code])
            pos += 1 + code

        elif code <= 78:
            # OP_PUSHDATA1, OP_PUSHDATA2, OP_PUSHDATA4, followed by length and data
            # push the data itself
            szsz = pow(2, code - 76)
            sz = encoding.decode(script[pos+szsz: pos:-1], 256)
            out.append(script[pos + 1 + szsz : pos + 1 + szsz + sz])
            pos += 1 + szsz + sz

        elif code <= 96:
            # OP_1NEGATE, OP_RESERVED, OP_1 thru OP_16
            # pass -1 for OP_1NEGATE
            # pass 0 for OP_RESERVED (shouldn't be used anyway)
            # pass 1 thru 16 for OP_1 thru OP_16
            out.append(code - 80)
            pos += 1

        else:
            # literal
            out.append(code)
            pos += 1

    # make sure each string is hex'ed
    out = encoding.json_changebase(out, lambda x: encoding.safe_hexlify(x))
    return out


def _btc_script_serialize_unit(unit):
    """
    Encode one item of a BTC script
    Return the encoded item (as a string)

    Based on code from pybitcointools (https://github.com/vbuterin/pybitcointools)
    by Vitalik Buterin
    """
    
    if isinstance(unit, int):
        if unit < 16:
            # OP_1 thru OP_16
            return encoding.from_int_to_byte(unit + 80)
        else:
            # pass literal
            return encoding.from_int_to_byte(unit)

    elif unit is None:
        # None means OP_0
        return b'\x00'

    else:
        if len(unit) <= 75:
            # length + payload
            return encoding.from_int_to_byte(len(unit)) + unit

        elif len(unit) < 256:
            # OP_PUSHDATA1 + length (1 byte) + payload
            return encoding.from_int_to_byte(OPCODE_VALUES['OP_PUSHDATA1']) + encoding.from_int_to_byte(len(unit)) + unit

        elif len(unit) < 65536:
            # OP_PUSHDATA2 + length (2 bytes, big-endian) + payload
            return encoding.from_int_to_byte(OPCODE_VALUES['OP_PUSHDATA2']) + encoding.encode(len(unit), 256, 2)[::-1] + unit
        else:
            # OP_PUSHDATA4 + length (4 bytes, big-endian) + payload
            return encoding.from_int_to_byte(OPCODE_VALUES['OP_PUSHDATA4']) + encoding.encode(len(unit), 256, 4)[::-1] + unit


def btc_script_serialize(_script):
    """
    Given a deserialized script (i.e. an array of Nones, ints, and strings), or an existing script,
    turn it back into a hex script

    Based on code from pybitcointools (https://github.com/vbuterin/pybitcointools)
    by Vitalik Buterin
    """
    script = _script
    if encoding.json_is_base(_script, 16):
        # hex-to-bin all hex strings in this script
        script = encoding.json_changebase(_script, lambda x: binascii.unhexlify(x))

    # encode each item and return the concatenated list
    return encoding.safe_hexlify( ''.join(map(_btc_script_serialize_unit, script)) )


def btc_make_payment_script( address, segwit=None, **ignored ):
    """
    Make a pay-to-address script.
    """
    
    if segwit is None:
        segwit = get_features('segwit')

    # is address bech32-encoded?
    witver, withash = segwit_addr_decode(address)
    if witver is not None and withash is not None:
        # bech32 segwit address
        if not segwit:
            raise ValueError("Segwit is disabled")

        if len(withash) == 20:
            # p2wpkh
            script_hex = '00' + withash.encode('hex')
            return script_hex
            
        elif len(withash) == 32:
            # p2wsh
            script_hex = '00' + withash.encode('hex')
            return script_hex

        else:
            raise ValueError("Unrecognized address '%s'" % address )

    else:
        # address is b58check-encoded
        vb = keylib.b58check.b58check_version_byte(address)
        if vb == version_byte:
            # p2pkh
            hash160 = binascii.hexlify( keylib.b58check.b58check_decode(address) )
            script = 'OP_DUP OP_HASH160 {} OP_EQUALVERIFY OP_CHECKSIG'.format(hash160)
            script_hex = btc_script_to_hex(script)
            return script_hex

        elif vb == multisig_version_byte:
            # p2sh
            hash160 = binascii.hexlify( keylib.b58check.b58check_decode(address) )
            script = 'OP_HASH160 {} OP_EQUAL'.format(hash160)
            script_hex = btc_script_to_hex(script)
            return script_hex

        else:
            raise ValueError("Unrecognized address '%s'" % address )


def btc_make_data_script( data, **ignored ):
    """
    Make a data-bearing transaction output.
    Data must be a hex string
    Returns a hex string.
    """
    if len(data) >= MAX_DATA_LEN * 2:
        raise ValueError("Data hex string is too long")     # note: data is a hex string

    if len(data) % 2 != 0:
        raise ValueError("Data hex string is not even length")

    return "6a{:02x}{}".format(len(data)/2, data)


def btc_script_hex_to_address( script_hex, segwit=None ):
    """
    Examine a script (hex-encoded) and extract an address.
    Return the address on success
    Return None on error
    """
    # TODO: make this support more than bitcoin-like scripts
    if script_hex.startswith("76a914") and script_hex.endswith("88ac") and len(script_hex) == 50:
        # p2pkh script
        hash160_bin = binascii.unhexlify(script_hex[6:-4])
        return bin_hash160_to_address(hash160_bin, version_byte=version_byte)

    elif script_hex.startswith("a914") and script_hex.endswith("87") and len(script_hex) == 46:
        # p2sh script
        hash160_bin = binascii.unhexlify(script_hex[4:-2])
        return bin_hash160_to_address(hash160_bin, version_byte=multisig_version_byte)

    elif script_hex.startswith('00') and len(script_hex) == 44:
        # p2wpkh script (bech32 address)
        hash160_bin = binascii.unhexlify(script_hex[2:])
        return segwit_addr_encode(hash160_bin) 

    elif script_hex.startswith('00') and len(script_hex) == 66:
        # p2wsh script (bech32 address)
        sha256_bin = binascii.unhexlify(script_hex[2:])
        return segwit_addr_encode(sha256_bin)

    return None


def btc_make_p2sh_address( script_hex ):
    """
    Make a P2SH address from a hex script
    """
    h = hashing.bin_hash160(binascii.unhexlify(script_hex))
    addr = bin_hash160_to_address(h, version_byte=multisig_version_byte)
    return addr


def btc_make_p2wpkh_address( pubkey_hex ):
    """
    Make a p2wpkh address from a hex pubkey
    """
    pubkey_hex = keylib.key_formatting.compress(pubkey_hex)
    hash160_bin = hashing.bin_hash160(pubkey_hex.decode('hex'))
    return segwit_addr_encode(hash160_bin)


def btc_make_p2sh_p2wpkh_redeem_script( pubkey_hex ):
    """
    Make the redeem script for a p2sh-p2wpkh witness script
    """
    pubkey_hash = hashing.bin_hash160(pubkey_hex.decode('hex')).encode('hex')
    redeem_script = btc_script_serialize(['0014' + pubkey_hash])
    return redeem_script


def btc_make_p2sh_p2wpkh_address( witness_script_hex ):
    """
    Make a p2sh address for a p2wpkh witness script hex
    """
    redeem_script = btc_make_p2sh_p2wpkh_redeem_script(witness_script_hex)
    p2sh_addr = btc_make_p2sh_address(redeem_script)
    return p2sh_addr


def btc_make_p2wsh_address( witness_script_hex ):
    """
    Make a p2wsh address from a witness script
    """
    witness_hash_bin = hashing.bin_sha256(witness_script_hex.decode('hex'))
    return segwit_addr_encode(witness_hash_bin)


def btc_make_p2sh_p2wsh_redeem_script( witness_script_hex ):
    """
    Make the redeem script for a p2sh-p2wsh witness script
    """
    witness_script_hash = hashing.bin_sha256(witness_script_hex.decode('hex')).encode('hex')
    redeem_script = btc_script_serialize(['0020' + witness_script_hash])
    return redeem_script


def btc_make_p2sh_p2wsh_address( witness_script_hex ):
    """
    Make a p2sh address for a p2wsh witness script hex
    """
    redeem_script = btc_make_p2sh_p2wsh_redeem_script(witness_script_hex)
    p2sh_addr = btc_make_p2sh_address(redeem_script)
    return p2sh_addr


def btc_is_p2sh_address( address ):
    """
    Is the given address a p2sh address?
    """
    vb = keylib.b58check.b58check_version_byte( address )
    if vb == multisig_version_byte:
        return True
    else:
        return False
    

def btc_is_p2pkh_address( address ):
    """
    Is the given address a p2pkh address?
    """
    vb = keylib.b58check.b58check_version_byte( address )
    if vb == version_byte:
        return True
    else:
        return False


def btc_is_p2wpkh_address( address ):
    """
    Is the given address a p2wpkh address?
    """
    wver, whash = segwit_addr_decode(address)
    if whash is None:
        return False

    if len(whash) != 20:
        return False

    return True


def btc_is_p2wsh_address( address ):
    """
    Is the given address a p2wsh address?
    """
    wver, whash = segwit_addr_decode(address)
    if whash is None:
        return False
    
    if len(whash) != 32:
        return False

    return True


def btc_is_segwit_address(address):
    """
    Is the given address a segwit (bech32) address?
    """
    return btc_is_p2wpkh_address(address) or btc_is_p2wsh_address(address)


def btc_is_p2sh_script( script_hex ):
    """
    Is the given scriptpubkey a p2sh script?
    """
    if script_hex.startswith("a914") and script_hex.endswith("87") and len(script_hex) == 46:
        return True
    else:
        return False


def btc_is_p2wsh_script( script_hex ):
    """
    Is the given scriptpubkey a p2wsh script?
    """
    if script_hex.startswith('00') and len(script_hex) == 66:
        return True
    else:
        return False


def btc_address_reencode( address, **blockchain_opts ):
    """
    Depending on whether or not we're in testnet 
    or mainnet, re-encode an address accordingly.
    """
    # re-encode bitcoin address
    network = blockchain_opts.get('network', None)
    opt_version_byte = blockchain_opts.get('version_byte', None)

    if btc_is_segwit_address(address):
        # bech32 address
        hrp = None
        if network == 'mainnet':
            hrp = 'bc'

        elif network == 'testnet':
            hrp = 'tb'

        else:
            if os.environ.get('BLOCKSTACK_TESTNET') == '1' or os.environ.get('BLOCKSTACK_TESTNET3') == '1':
                hrp = 'tb'

            else:
                hrp = 'bc'

        wver, whash = segwit_addr_decode(address)
        return segwit_addr_encode(whash, hrp=hrp, witver=wver)

    else:
        # base58 address
        vb = keylib.b58check.b58check_version_byte( address )

        if network == 'mainnet':
            if vb == 0 or vb == 111:
                vb = 0

            elif vb == 5 or vb == 196:
                vb = 5

            else:
                raise ValueError("Unrecognized address %s" % address)
        
        elif network == 'testnet':
            if vb == 0 or vb == 111:
                vb = 111
            
            elif vb == 5 or vb == 196:
                vb = 196

            else:
                raise ValueError("Unrecognized address %s" % address)

        else:
            if opt_version_byte is not None:
                vb = opt_version_byte

            elif os.environ.get("BLOCKSTACK_TESTNET") == "1" or os.environ.get("BLOCKSTACK_TESTNET3") == "1":
                if vb == 0 or vb == 111:
                    # convert to testnet p2pkh
                    vb = 111

                elif vb == 5 or vb == 196:
                    # convert to testnet p2sh
                    vb = 196

                else:
                    raise ValueError("unrecognized address %s" % address)

            else:
                if vb == 0 or vb == 111:
                    # convert to mainnet p2pkh
                    vb = 0

                elif vb == 5 or vb == 196:
                    # convert to mainnet p2sh
                    vb = 5

                else:
                    raise ValueError("unrecognized address %s" % address)

        return keylib.b58check.b58check_encode( keylib.b58check.b58check_decode(address), vb )


def btc_is_multisig(privkey_info, **blockchain_opts):
    """
    Does the given private key info represent
    a multisig bundle?

    For Bitcoin, this is true for multisig p2sh (not p2sh-p2wsh)
    """
    try:
        jsonschema.validate(privkey_info, PRIVKEY_MULTISIG_SCHEMA)
        return not privkey_info.get('segwit', False)
    except ValidationError as e:
        return False


def btc_is_multisig_segwit(privkey_info):
    """
    Does the given private key info represent
    a multisig bundle?

    For Bitcoin, this is true for multisig p2sh (not p2sh-p2wsh)
    """
    try:
        jsonschema.validate(privkey_info, PRIVKEY_MULTISIG_SCHEMA)
        if len(privkey_info['private_keys']) == 1:
            return False

        return privkey_info.get('segwit', False)
    except ValidationError as e:
        return False


def btc_is_multisig_address(addr, **blockchain_opts):
    """
    Is the given address a multisig address?
    """
    return btc_is_p2sh_address(addr) or btc_is_p2wsh_address(addr)


def btc_is_multisig_script(script_hex, **blockchain_opts):
    """
    Is the given script hex a multisig script?
    """
    return btc_is_p2sh_script(script_hex) or btc_is_p2wsh_script(script_hex)


def btc_is_singlesig(privkey_info, **blockchain_opts):
    """
    Does the given private key info represent
    a single signature bundle? (i.e. one private key)?

    i.e. is this key a private key string?
    """
    try:
        jsonschema.validate(privkey_info, PRIVKEY_SINGLESIG_SCHEMA)
        return True
    except ValidationError as e:
        return False


def btc_get_singlesig_privkey(privkey_info, **blockchain_opts):
    """
    Get the single-sig private key from the private key info
    """
    if btc_is_singlesig(privkey_info):
        return privkey_info

    elif btc_is_singlesig_segwit(privkey_info):
        return privkey_info['private_keys'][0]

    return None


def btc_is_singlesig_address(addr, **blockchain_opts):
    """
    Is the given address a single-sig address?
    """
    return btc_is_p2pkh_address(addr)


def btc_is_singlesig_segwit(privkey_info):
    """
    Is the given key bundle a p2sh-p2wpkh key bundle?
    """
    try:
        jsonschema.validate(privkey_info, PRIVKEY_MULTISIG_SCHEMA)
        if len(privkey_info['private_keys']) > 1:
            return False

        return privkey_info.get('segwit', False)
    except ValidationError:
        return False


def btc_get_privkey_address(privkey_info, **blockchain_opts):
    """
    Get the address for a given private key info bundle
    (be it multisig or singlesig)

    Return the address on success
    Raise exception on error
    """
    
    from .multisig import make_multisig_segwit_address_from_witness_script

    if btc_is_singlesig(privkey_info):
        return btc_address_reencode( ecdsalib.ecdsa_private_key(privkey_info).public_key().address() )
    
    if btc_is_multisig(privkey_info) or btc_is_singlesig_segwit(privkey_info):
        redeem_script = str(privkey_info['redeem_script'])
        return btc_make_p2sh_address(redeem_script)
    
    if btc_is_multisig_segwit(privkey_info):
        return make_multisig_segwit_address_from_witness_script(str(privkey_info['redeem_script']))

    raise ValueError("Invalid private key info")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    Virtualchain
    ~~~~~
    copyright: (c) 2014 by Halfmoon Labs, Inc.
    copyright: (c) 2015 by Blockstack.org
    
    This file is part of Virtualchain
    
    Virtualchain is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    
    Virtualchain is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with Virtualchain.  If not, see <http://www.gnu.org/licenses/>.
"""

import logging
import os
import httplib
import ssl
import socket

from ..config import get_logger, get_bitcoind_config

log = get_logger("virtualchain_session")

# various SSL compat measures
create_ssl_authproxy = False 
do_wrap_socket = False

if hasattr( ssl, "_create_unverified_context" ):
   ssl._create_default_https_context = ssl._create_unverified_context
   create_ssl_authproxy = True 

if not hasattr( ssl, "create_default_context" ):
   create_ssl_authproxy = False
   do_wrap_socket = True


# disable debug logging from bitcoinrpc
bitcoinrpc_logger = logging.getLogger("BitcoinRPC")
bitcoinrpc_logger.setLevel(logging.CRITICAL)

class BitcoindConnection( httplib.HTTPSConnection ):
   """
   Wrapped SSL connection, if we can't use SSLContext.
   """

   def __init__(self, host, port, timeout=None ):
   
      httplib.HTTPSConnection.__init__(self, host, port )
      self.timeout = timeout
        
   def connect( self ):
      
      sock = socket.create_connection((self.host, self.port), self.timeout)
      if self._tunnel_host:
         self.sock = sock
         self._tunnel()
         
      self.sock = ssl.wrap_socket( sock, cert_reqs=ssl.CERT_NONE )


def is_int(i):
    """
    Is the given object a long or an int?
    """
    return isinstance(i, (int,long))


def is_valid_int(i):
    """
    Is the given object an integer?
    """
    if is_int(i):
        return True
    elif isinstance(i, str):
        try:
            int_i = int(i)
        except:
            return False
        else:
            return True
    return False


def create_bitcoind_connection( rpc_username, rpc_password, server, port, use_https, timeout ):
    """
    Creates an RPC client to a bitcoind instance.
    It will have ".opts" defined as a member, which will be a dict that stores the above connection options.
    """
    
    from .bitcoin_blockchain import AuthServiceProxy

    global do_wrap_socket, create_ssl_authproxy
        
    log.debug("[%s] Connect to bitcoind at %s://%s@%s:%s, timeout=%s" % (os.getpid(), 'https' if use_https else 'http', rpc_username, server, port, timeout) )
    
    protocol = 'https' if use_https else 'http'
    if not server or len(server) < 1:
        raise Exception('Invalid bitcoind host address.')
    if not port or not is_valid_int(port):
        raise Exception('Invalid bitcoind port number.')
    
    authproxy_config_uri = '%s://%s:%s@%s:%s' % (protocol, rpc_username, rpc_password, server, port)
    
    if use_https:
        # TODO: ship with a cert
        if do_wrap_socket:
           # ssl._create_unverified_context and ssl.create_default_context are not supported.
           # wrap the socket directly 
           connection = BitcoindConnection( server, int(port), timeout=timeout )
           ret = AuthServiceProxy(authproxy_config_uri, connection=connection)
           
        elif create_ssl_authproxy:
           # ssl has _create_unverified_context, so we're good to go 
           ret = AuthServiceProxy(authproxy_config_uri, timeout=timeout)
        
        else:
           # have to set up an unverified context ourselves 
           ssl_ctx = ssl.create_default_context()
           ssl_ctx.check_hostname = False
           ssl_ctx.verify_mode = ssl.CERT_NONE
           connection = httplib.HTTPSConnection( server, int(port), context=ssl_ctx, timeout=timeout )
           ret = AuthServiceProxy(authproxy_config_uri, connection=connection)
          
    else:
        ret = AuthServiceProxy(authproxy_config_uri)

    # remember the options 
    bitcoind_opts = {
       "bitcoind_user": rpc_username,
       "bitcoind_passwd": rpc_password,
       "bitcoind_server": server,
       "bitcoind_port": port,
       "bitcoind_use_https": use_https,
       "bitcoind_timeout": timeout
    }
    
    setattr( ret, "opts", bitcoind_opts )
    return ret


def connect_bitcoind_impl( bitcoind_opts ):
    """
    Create a connection to bitcoind, using a dict of config options.
    """

    if 'bitcoind_port' in bitcoind_opts.keys() and bitcoind_opts['bitcoind_port'] is None:
        log.error("No port given")
        raise ValueError("No RPC port given (bitcoind_port)")

    if 'bitcoind_timeout' in bitcoind_opts.keys() and bitcoind_opts['bitcoind_timeout'] is None:
        # default
        bitcoind_opts['bitcoind_timeout'] = 300

    try:
        int(bitcoind_opts['bitcoind_port'])
    except:
        log.error("Not an int: '%s'" % bitcoind_opts.get('bitcoind_port'))
        raise

    try:
        float(bitcoind_opts.get('bitcoind_timeout', 300))
    except:
        log.error("Not a float: '%s'" % bitcoind_opts.get('bitcoind_timeout', 300))
        raise

    return create_bitcoind_connection( bitcoind_opts['bitcoind_user'], bitcoind_opts['bitcoind_passwd'], \
                                       bitcoind_opts['bitcoind_server'], int(bitcoind_opts['bitcoind_port']), \
                                       bitcoind_opts.get('bitcoind_use_https', False), float(bitcoind_opts.get('bitcoind_timeout', 300)) )
 

def get_bitcoind_client(config_path=None, bitcoind_opts=None):
    """
    Connect to bitcoind
    """
    if bitcoind_opts is None and config_path is None:
        raise ValueError("Need bitcoind opts or config path")

    bitcoind_opts = get_bitcoind_config(config_file=config_path)
    log.debug("Connect to bitcoind at %s:%s (%s)" % (bitcoind_opts['bitcoind_server'], bitcoind_opts['bitcoind_port'], config_path))
    client = connect_bitcoind_impl( bitcoind_opts )

    return client
 





#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    Virtualchain
    ~~~~~
    copyright: (c) 2014 by Halfmoon Labs, Inc.
    copyright: (c) 2015-2018 by Blockstack.org
    
    This file is part of Virtualchain
    
    Virtualchain is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    
    Virtualchain is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with Virtualchain.  If not, see <http://www.gnu.org/licenses/>.
"""

from bitcoin_blockchain import BlockchainDownloader, SPVClient

import os
import time
import random
from decimal import *

from ..config import get_logger

log = get_logger("virtualchain")

from bitcoin_blockchain.bits import *
from bitcoin_blockchain.blocks import get_bitcoin_virtual_transactions

def get_virtual_transactions(blockchain_name, blockchain_opts, first_block_height, last_block_height, tx_filter=None, **hints):
    """
    Get the sequence of virtualchain transactions from a particular blockchain over a given range of block heights.
    Returns a list of tuples in the format of [(block height, [txs])], where
    each tx in [txs] is the parsed transaction.  The parsed transaction will conform to... # TODO write a spec for this
    
    Each transaction has at least the following fields:
    
    `version`: the version of the transaction
    `txindex`: the index into the block where this tx occurs
    `ins`: a list of transaction inputs, where each member is a dict with:
        `outpoint`: a dict of {'hash': txid of transaction that fed it in, 'index': the index into the feeder tx's outputs list}
        `script`: the signature script for this input
    `outs`: a list of transaction outputs, where each member is a dict with:
        `value`: the amount of currency units spent (in the fundamental units of the chain)
        `script`: the spending script for this input
    `senders`: a list of information in 1-to-1 correspondence with each input regarding the transactions that funded it:
        `value`: the amount of currency units sent (in fundamental units of the chain) 
        `script_pubkey`: the spending script for the sending transaction
    
    Returns [(block height, [txs])] on success
    Returns None on error.
    Raises ValueError on unknown blockchain
    """
    if blockchain_name == 'bitcoin':
        return get_bitcoin_virtual_transactions(blockchain_opts, first_block_height, last_block_height, tx_filter=tx_filter, **hints)

    else:
        raise ValueError("Unknown blockchain {}".format(blockchain_name))


def tx_parse(raw_tx, blockchain='bitcoin', **blockchain_opts):
    """
    Parse a raw transaction, based on the type of blockchain it's from
    Returns a tx dict on success (see get_virtual_transactions)
    Raise ValueError for unknown blockchain
    Raise some other exception for invalid raw_tx (implementation-specific)
    """
    if blockchain == 'bitcoin':
        return btc_tx_deserialize(raw_tx, **blockchain_opts)
    else:
        raise ValueError("Unknown blockchain {}".format(blockchain))


def tx_output_has_data(output, blockchain='bitcoin', **blockchain_opts):
    """
    Give a blockchain name and a tx output, determine whether or not it is a
    data-bearing script--i.e. one with data for the state engine.

    Return True if so
    Return False if not
    """
    if blockchain == 'bitcoin':
        return btc_tx_output_has_data(output, **blockchain_opts)
    else:
        return ValueError('Unknown blockchain "{}"'.format(blockchain))
    

def tx_is_data_script(out_script, blockchain='bitcoin', **blockchain_opts):
    """
    Given a blockchain name and an output script (tx['outs'][x]['script']),
    determine whether or not it is a data-bearing script---i.e. one with data for the state engine.

    Return True if so
    Reurn False if not
    """
    if blockchain == 'bitcoin':
        return btc_tx_output_script_has_data(out_script, **blockchain_opts)
    else:
        raise ValueError('Unknown blockchain "{}"'.format(blockchain))


def tx_extend(partial_tx_hex, new_inputs, new_outputs, blockchain='bitcoin', **blockchain_opts):
    """
    Add a set of inputs and outputs to a tx.
    Return the new tx on success
    Raise on error
    """
    if blockchain == 'bitcoin':
        return btc_tx_extend(partial_tx_hex, new_inputs, new_outputs, **blockchain_opts)
    else:
        raise ValueError('Unknown blockchain "{}"'.format(blockchain))


def tx_sign_input(tx_hex, idx, prevout_script, prevout_amount, private_key_info, blockchain='bitcoin', **blockchain_opts):
    """
    Sign a given input in a transaction, given the previous output script and previous output amount.
    Different blockchains can require additional fields; pass thse in **blockchain_opts.

    Return the serialized tx with the given input signed on success
    Raise on error
    """
    if blockchain == 'bitcoin':
        return btc_tx_sign_input(tx_hex, idx, prevout_script, prevout_amount, private_key_info, **blockchain_opts)
    else:
        raise ValueError('Unknown blockchain "{}"'.format(blockchain))


def tx_sign_all_unsigned_inputs(privkey_info, prev_outputs, unsigned_tx_hex, blockchain='bitcoin', **blockchain_opts):
    """
    Sign all unsigned inputs to a given transaction with the given private key.  Also, pass in the list of previous outputs to the transaction so they
    can be paired with the right input (i.e. prev_outputs is a list of tx outputs that are in 1-to-1 correspondance with the inputs in the serialized tx)
    
    Different blockchains can require additional fields; pass thse in **blockchain_opts.

    Return the signed transaction on success
    Raise on error
    """
    if blockchain == 'bitcoin':
        return btc_tx_sign_all_unsigned_inputs(privkey_info, prev_outputs, unsigned_tx_hex, **blockchain_opts)
    else:
        raise ValueError('Unknown blockchain "{}"'.format(blockchain))
    

def calculate_change_amount(inputs, send_amount, fee):
    """
    Find out how much change there exists between a set of tx inputs, the send amount and the tx fee.
    @inputs must be a list of transaction inputs, i.e. [{'script_hex': str, 'value': int}]
    """
    # calculate the total amount coming into the transaction from the inputs
    total_amount_in = sum([input['value'] for input in inputs])

    # change = whatever is left over from the amount sent & the transaction fee
    change_amount = total_amount_in - send_amount - fee

    # check to ensure the change amount is a non-negative value and return it
    if change_amount < 0:
        raise ValueError('Not enough inputs for transaction (total: {}, to spend: {}, fee: {}).'.format(total_amount_in, send_amount, fee))

    return change_amount





#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    Virtualchain
    ~~~~~
    copyright: (c) 2014-15 by Halfmoon Labs, Inc.
    copyright: (c) 2016 by Blockstack.org

    This file is part of Virtualchain

    Virtualchain is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Virtualchain is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with Virtualchain. If not, see <http://www.gnu.org/licenses/>.
"""

from setuptools import setup, find_packages

# to set __version__
exec(open('virtualchain/version.py').read())

setup(
    name='virtualchain',
    version=__version__,
    url='https://github.com/blockstack/virtualchain',
    license='GPLv3',
    author='Blockstack.org',
    author_email='support@blockstack.org',
    description='A library for constructing virtual blockchains within a cryptocurrency\'s blockchain',
    keywords='blockchain bitcoin btc cryptocurrency data',
    packages=find_packages(),
    download_url='https://github.com/blockstack/virtualchain/archive/master.zip',
    zip_safe=False,
    include_package_data=True,
    install_requires=[
        'protocoin>=0.2',
        'simplejson>=3.8.2',
        'jsonschema>=2.5.1',
        'cryptography>=1.9',
        'ecdsa>=0.13',
        'six>=1.10.0',
        'keylib>=0.1.1',
    ],
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet',
        'Topic :: Security :: Cryptography',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    Virtualchain
    ~~~~~
    copyright: (c) 2014-2015 by Halfmoon Labs, Inc.
    copyright: (c) 2016 by Blockstack.org
    
    This file is part of Virtualchain
    
    Virtualchain is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    
    Virtualchain is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with Virtualchain.  If not, see <http://www.gnu.org/licenses/>.
"""

from .keys import BitcoinPublicKey, BitcoinPrivateKey, hex_hash160_to_address, btc_script_hex_to_address, version_byte, btc_is_multisig_script, \
        btc_make_payment_script, btc_make_data_script, btc_address_reencode, btc_is_p2sh_script, btc_is_p2sh_address, MAX_DATA_LEN

from .fees import get_tx_fee_per_byte, get_tx_fee, tx_estimate_signature_len

from .multisig import *
from .authproxy import *

from .spv import SPVClient
from .blocks import BlockchainDownloader, get_bitcoin_blockchain_height
from .bits import *
